# -*- coding: utf-8 -*-
"""
Created on Sat Oct  7 10:44:36 2017

@author: mwojc
"""
from fempy.compat import builtins
import inspect
import numpy as np
import scipy.sparse
import scipy.optimize
from fempy.solvers import precon
from fempy.solvers import methods
from fempy.solvers._solver import Solver

LOP = scipy.sparse.linalg.LinearOperator


def hasprop(obj, name):
    obj = obj.__class__ if not inspect.isclass(obj) else obj
    if builtins.hasattr(obj, name):
        if isinstance(getattr(obj, name), property):
            return True
    return False


def hasattr(obj, name):
#    # first test for class properties...
#    if __builtin__.hasattr(obj.__class__, name):
#        if isinstance(getattr(obj.__class__, name), property):
#            return True
#    # and then test object...
#    return __builtin__.hasattr(obj, name)
    return hasprop(obj, name) or builtins.hasattr(obj, name)


def _append(*args):
    return np.concatenate([a for a in args if a is not None])


def _initialize(obj, msg='obj'):
    if inspect.isclass(obj):
        try:
            obj = obj()
        except Exception:
            raise TypeError("Please, initialize {}...".format(msg))
    return obj


def _set_solution(m, x):
    if hasattr(m, 'lhslm'):
        nlm = m.lhslm.size
        xlm = x[-nlm:]
        if hasprop(m, 'lhslm'):
            m.lhslm = xlm
        else:
            m.lhslm[:] = xlm
        x = x[:-nlm]
    if hasprop(m, 'lhs'):
        m.lhs = x
    else:
        m.lhs[:] = x.reshape(m.lhs.shape)


def _get_jac_times_lhs(model):
    # jac is always 2D?
    lhs = model.lhs
    r = lhs.copy()  # return same type as lhs
    r[:] = model.jac.dot(lhs.ravel()).reshape(lhs.shape)
    return r


def _get_loplm(model):
    # jaclm is always 2D?
    lhs = model.lhs
    lhslm = model.lhslm
    jaclm = model.jaclm
    rx = lhs.copy()  # return same type as lhs
    rx[:] = jaclm.T.dot(lhslm).reshape(lhs.shape)
    rlm = jaclm.dot(lhs.ravel())
    return rx, rlm


def _get_res(model):
    r = model.lop
    rhs = model.rhs
    if not hasattr(model, 'jaclm'):
        return r - rhs
    rx, rlm = model.loplm
    rhslm = model.rhslm
    return r + rx - rhs, rlm - rhslm


def _get_lop(model):
    r = model.lop
    if not hasattr(model, 'jaclm'):
        return r
    rx, rlm = model.loplm
    return r + rx, rlm


def _solvelm(P, B, BT, Mlm, bx, blm):
    blm = (B.dot(P.dot(bx)).T - blm).T
    lm = Mlm.dot(blm)  # get lagrange multipliers
    x = P.dot(bx - BT.dot(lm))
    return x, lm


def _unroll_bcseq(m):
    bcseq = m.bcseq
    bcseqm = []
    bcseqs = []
    bcseqc = []
    for eq in bcseq:
        n = len(eq)
        coeffs = (1., -1., 0.) if n < 3 else eq[2]
        mr = eq[0]  # master
        sv = eq[1]  # slave
        mr = (mr,) if not isinstance(mr, tuple) else mr
        sv = (sv,) if not isinstance(sv, tuple) else sv
        # mr = list(mr) if not isinstance(mr, (int, str)) else [mr]
        # sv = list(sv) if not isinstance(sv, (int, str)) else [sv]
        if hasattr(m, 'domain'):
            sort = None if n < 4 else eq[3]
            # get properly sorted indices
            if isinstance(mr[0], str):
                mr = (m.domain.ngroups.nums(mr[0], sortby=sort),) + mr[1:]
            if isinstance(sv[0], str):
                sv = (m.domain.ngroups.nums(sv[0], sortby=sort),) + sv[1:]
            # get global numbers of dofs
            mr = m.lhs.nipos[mr].ravel()
            sv = m.lhs.nipos[sv].ravel()
            mr = np.asarray([mr]) if isinstance(mr, int) else mr
            sv = np.asarray([sv]) if isinstance(sv, int) else sv
        N = len(mr)
        assert N == len(sv)
        bcseqm += [mr]
        bcseqs += [sv]
#        bcseqc += [np.tile(coeffs, (len(mr), 1))]
        c0, c1, c2 = coeffs
        c0 = np.zeros(N)+c0 if isinstance(c0, (float, int)) else \
            np.asarray(c0).ravel()
        c1 = np.zeros(N)+c1 if isinstance(c1, (float, int)) else \
            np.asarray(c1).ravel()
        c2 = np.zeros(N)+c2 if isinstance(c2, (float, int)) else \
            np.asarray(c2).ravel()
        bcseqc += [np.asarray([c0, c1, c2]).T]
    npc = np.concatenate
    bcseq0, bcseq1, bcseqc = npc(bcseqm), npc(bcseqs), npc(bcseqc).T
#    return npc(bcseqm), npc(bcseqs), npc(bcseqc).T
#    # Perform some tests for duplicated constraints
#    # find duplicates in bcseq - only first occurence is left, master nodes
#    # are considered first, prescribed msters are protected
#    from fempy.domain.tools import unique_rows
#    nb = len(bcseq0)
#    uidx = unique_rows(npc((bcseq0, bcseq1))[:, None], strict=False)
#    uidx = uidx[:nb]*uidx[nb:] #+ ibm  # protect prescribed nodes
#    bcseq0 = bcseq0[uidx]
#    bcseq1 = bcseq1[uidx]
#    bcseqc = bcseqc[:, uidx]
    # find conflicting bcs and bcseq constraints
    if hasattr(m, 'bcs'):
        bcsr = m.bcs.ravel()
        lhsr = m.lhs.ravel()
        # 'slave' unknown is prescribed - reverse order
        b0 = bcsr[bcseq0]
        b1 = bcsr[bcseq1]
        ibs = ~b0 * b1
        bcseq0[ibs], bcseq1[ibs] = bcseq1[ibs], bcseq0[ibs]
        bcseqc[0, ibs], bcseqc[1, ibs] = bcseqc[1, ibs], bcseqc[0, ibs]
        # 'master' unknown is prescribed, TODO: remove from bcseq ???
        b0 = bcsr[bcseq0]
        b1 = bcsr[bcseq1]
        ibm = b0 * ~b1
        bcseqc[2, ibm] = bcseqc[2, ibm] - bcseqc[0, ibm] * lhsr[bcseq0[ibm]]
        bcseqc[0, ibm] = 0.
        # both unknowns are prescribed
        ib = b0 * b1
        bcseq0 = bcseq0[~ib]
        bcseq1 = bcseq1[~ib]
        bcseqc = bcseqc[..., ~ib]
        ibm = ibm[~ib]
    # find duplicates in bcseq - only first occurence is taken if dup
    from fempy.domain.tools import unique_rows
    beq = np.concatenate(([bcseq0], [bcseq1])).T
    idx = unique_rows(beq, sort=True, strict=False)
    idx *= ~(bcseq0 == bcseq1)  # exclude also "self-periodicity"
    bcseq0, bcseq1 = beq[idx].T
    bcseqc = bcseqc[..., idx]
    # Masters cannot be duplicated (unless they are prescribed)
    ibm = ibm[idx]
    midx = unique_rows(bcseq0[:, None], strict=False) + ibm
    bcseq0 = bcseq0[midx]
    bcseq1 = bcseq1[midx]
    bcseqc = bcseqc[:, midx]
    # Slaves cannot be duplicated
    ibm = ibm[midx]
    sidx = unique_rows(bcseq1[:, None], strict=False)
    bcseq0 = bcseq0[sidx]
    bcseq1 = bcseq1[sidx]
    bcseqc = bcseqc[:, sidx]
#    # Slaves cannot be masters (unless corresponding masters are prescribed)
    ibm = ibm[sidx]
    sidx = ~np.isin(bcseq1, bcseq0) + ibm
    bcseq0 = bcseq0[sidx]
    bcseq1 = bcseq1[sidx]
    bcseqc = bcseqc[:, sidx]
    return bcseq0, bcseq1, bcseqc


def _multiply(A, b):
    if not scipy.sparse.issparse(A):
        return A * b
    else:
        return A.multiply(b)


# We derive from Solver only for proper options handling
# Defaut jacobian uses direct linear solver (superlu from scipy)
class LinearSolver(Solver):
    default_preconditioner = precon.exact
    default_method = methods.direct

    def __init__(self, model, **options):
        # Assign and update model
        self.model = _initialize(model, 'model')
#        if hasattr(self.model, 'update'):
#            self.model.update()
        # Query model for all necessary data
        m = self.model
        if m is None:
            return
        # test for lop or jac
        if not hasattr(m, 'lop') and not hasattr(m, 'jac'):
            raise AttributeError("Neither 'jac' nor 'lop' found in model")
        # now we query for lhs
        if not hasattr(m, 'lhs'):
            if hasattr(m, 'rhs'):
                m.lhs = np.zeros_like(m.rhs)
            else:
                if hasattr(m, 'lop'):
                    m.lhs = np.zeros_like(m.lop)
                else:
                    jac = m.jac
                    m.lhs = np.zeros((jac.shape[0],), jac.dtype)
        self.nlhs = m.lhs.size
        # now jac and lop
        if not hasattr(m, 'lop'):   # only jac is given
            m.__class__.lop = property(_get_jac_times_lhs)
        if not hasattr(m, 'jac'):  # only lop is given
            m.jac = None  # solvers have to decide what to do without jac
        # create jac testing utility
        if not hasattr(m, 'jac_times_lhs'):
            m.__class__.jac_times_lhs = property(_get_jac_times_lhs)
        #  we test for rhs also
        if not hasattr(m, 'rhs'):
            m.rhs = np.zeros_like(m.lhs)
        # bcs is optional, we create it here, but we
        # should not rely on its existence in solvers
        if not hasattr(m, 'bcs'):
            m.bcs = np.zeros_like(m.lhs, dtype=bool)
        # now lagrange multipliers part, created only if jaclm exists in model
        if hasattr(m, 'jaclm'):
            if not hasattr(m, 'rhslm'):
                m.rhslm = np.zeros(len(m.jaclm))
            if not hasattr(m, 'lhslm'):
                m.lhslm = np.zeros(len(m.jaclm))
            if not hasattr(m, 'loplm'):
                m.__class__.loplm = property(_get_loplm)
            self.nlm = len(m.lhslm)
        else:
            self.nlm = 0
        # create also residuals, this will be a tuple if Lagrange
        # multipliers are defined
        if not hasattr(m, 'res'):
            m.__class__.res = property(_get_res)
        # and also set time attributes
        if not hasattr(m, 'time'):
            m.time = 0.
        if not hasattr(m, 'time0'):
            m.time0 = 0.
        if not hasattr(m, 'time1'):
            m.time1 = 1.
        # now set preconditioner...
        if 'preconditioner' in options:
            preconditioner = options['preconditioner']
            del options['preconditioner']
        else:
            preconditioner = 'default'
        self.preconditioner = preconditioner
        # ... and linear solver
        if 'method' in options:
            method = options['method']
            del options['method']
        else:
            method = 'default'
        self.method = method
#        # ... and options
#        # ... recognize and set linear solver options
#        for key, value in options.items():
#            if key.startswith('inner_'):
#                self.method.options[key[6:]] = value
        self.options = options
        # update finally
        # self.update()

    @property
    def preconditioner(self):
        return self._preconditioner

    @preconditioner.setter
    def preconditioner(self, p):
        if p == 'default':
            p = _initialize(self.default_preconditioner)
        self._preconditioner = _initialize(p, 'preconditioner')

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, l):
        if l == 'default':
            l = _initialize(self.default_method)
        self._method = _initialize(l, 'linear solver')

    @property
    def x(self):
        m = self.model
        # query for lhs ...
        lhs = m.lhs
        self.nlhs = lhs.size
        self.shape_lhs = lhs.shape
        self.dtype = lhs.dtype
        lhs = lhs.copy().ravel()
        # for lhslm ...
        if not hasattr(m, 'lhslm'):
            self.nlm = 0
            x = lhs
        else:
            lhslm = m.lhslm
            self.nlm = len(lhslm)
            x = _append(lhs, lhslm)
        # for bcs ...
        if hasattr(m, 'bcs'):
            bcs = m.bcs.ravel()
            self.nbcs = bcs.sum()
            self.bcsr = bcs.copy()
            self.x_bcsr = lhs[bcs].copy()
        else:
            self.nbcs = 0
            self.bcsr = np.zeros_like(lhs, bool)  # always have bcsr
        # and bcseq
        if hasattr(m, 'bcseq'):
            self.bcseq0, self.bcseq1, self.bcseqc = _unroll_bcseq(m)
            self.x_bcseq0 = lhs[self.bcseq0].copy()
            self.x_bcseq1 = lhs[self.bcseq1].copy()
            self.nbcseq = len(self.bcseq0)
        else:
            self.nbcseq = 0
        self.n = len(x)
        self.shape = (self.n, )*2
        return x

    @x.setter
    def x(self, x):
        if x is None:
            return
        _set_solution(self.model, x)

    @property
    def b(self):
        # x must be called first!
        m = self.model
        rhs = m.rhs.copy().ravel()
        if self.nbcseq:
            c0, c1, c2 = self.bcseqc
            rhs[self.bcseq0] += rhs[self.bcseq1]
            rhs[self.bcseq1] = c2/c1
        if self.nbcs:
            rhs[self.bcsr] = self.x_bcsr
        b = rhs if not self.nlm else _append(rhs, m.rhslm)
        return b

    @property
    def l(self):
        # x must be called first! !!!
        m = self.model
        lx = m.lop.ravel()
        llm = None
        if self.nlm:
            llmx, llm = m.loplm
            lx += llmx.ravel()
        if self.nbcseq:
            c0, c1, c2 = self.bcseqc
            lx[self.bcseq0] += lx[self.bcseq1]
            lx[self.bcseq1] = (c0/c1)*self.x_bcseq0 + self.x_bcseq1  # proper x
        if self.nbcs:
            lx[self.bcsr] = self.x_bcsr
        return _append(lx, llm)

    @property
    def r(self):
        return self.l - self.b

    def _update_K(self):
        self.K = self.model.jac

    def _update_B(self):
        if self.nlm:
            self.B = self.model.jaclm
            # make sure B is zero at bcsr
            # this is done here because it is necessary in _update_Mlm and
            # _matvecJ
            if self.nbcs:
                self.B[:, self.bcsr] = 0.
            self.BT = self.B.T
        else:
            self.B = None
            self.BT = None

    def _update_P(self):
        K = self.K
        nbcs = self.nbcs
        nbcseq = self.nbcseq
        pre = self.preconditioner if hasattr(self, 'preconditioner') else None
        if K is None or pre is None:
            self.P = None
            return
        if not nbcs and not nbcseq:
            P = pre(K)  # LOP or sp matrix
        else:
            bcsr = self.bcsr.copy()
            if nbcseq:
                K = K.copy()  # this is safe but this is a copy...
                # TODO: K should not be copied here at all
                c0, c1, c2 = self.bcseqc
                bcseq0 = self.bcseq0
                bcseq1 = self.bcseq1
                c01 = c0/c1
                K[bcseq0] += K[bcseq1]
                K[:, bcseq0] -= _multiply(K[:, bcseq1], c01)  # elementwise
                bcsr[bcseq1] = True
            K0 = K[~bcsr]
            K0 = K0[:, ~bcsr]
            Kbcs = K[:, bcsr]
            P0 = pre(K0)  # LOP or sp matrix

            def matvec(b):
                bbcs = (Kbcs.dot(b[bcsr]))[~bcsr]  # b carries x at bcsr
                b0 = b[~bcsr] - bbcs
                x0 = P0.dot(b0)
                x = b.copy()                     # copies also x at bcsr
                x[~bcsr] = x0
                if nbcseq:
                    x[bcseq1] -= c01 * x[bcseq0]
                return x
            P = LOP(K.shape, dtype=K.dtype, matvec=matvec)  # LOP
        self.P = P

    def _update_Klm(self):
        P = self.P
        if P is None:
            self.Klm = None
            return
        B = self.B
        BT = self.BT
        if B is None:
            self.Klm = None
            return
        self.Klm = B.dot(P.dot(BT))      # Use preconditioner to inverse?

    def _update_Mlm(self):
        Klm = self.Klm
        if Klm is None:
            self.Mlm = None
            return
        if scipy.sparse.issparse(Klm):   # Efficiency for many lms??
            Mlm = scipy.sparse.linalg.inv(Klm)
        else:
            Mlm = np.linalg.inv(Klm)
        self.Mlm = Mlm

    def _matvecJ(self, x):
        K = self.K
        if K is None:
            # then we do not update nothing in solver, just in model
            self.x = x
            self.update_model()  # or simply update :)?
            self.x  # l uses x_bcsr so we need to update...
            return self.l  # self.r + self.b
            # return self.residual(x)
        if not self.nlm:
            lx = K.dot(x)
            llm = None
        else:
            xx, xlm = x[:self.nlhs], x[self.nlhs:]
            B = self.B
            BT = self.BT
            lx = K.dot(xx) + BT.dot(xlm)
            llm = B.dot(xx)
        # TODO: the below is OK for symmetric K only
        # Consider jacobian manipulation here
        if self.nbcseq:
            c0, c1, c2 = self.bcseqc
            lx[self.bcseq0] += lx[self.bcseq1]  # add rows
            x_bcseq1 = x[:self.nlhs][self.bcseq1]
            x_bcseq0 = x[:self.nlhs][self.bcseq0]
            lx[self.bcseq1] = (c0/c1)*x_bcseq0 + x_bcseq1  # proper x
        if self.nbcs:
            x_bcsr = x[:self.nlhs][self.bcsr]
            lx[self.bcsr] = x_bcsr
        return _append(lx, llm)

    def _matvecM(self, b):
        P = self.P
        if not self.nlm:
            x = P.dot(b)
            lm = None
        else:
            bx, blm = b[:self.nlhs], b[self.nlhs:]
            B = self.B
            BT = self.BT
            Mlm = self.Mlm
            x, lm = _solvelm(P, B, BT, Mlm, bx, blm)
#        if self.nbcs:  # This is done in P and Mlm already...
#            x[self.bcsr] = b[self.bcsr]
        return _append(x, lm)

    @property
    def J(self):
        return LOP(self.shape, matvec=self._matvecJ)

    @property
    def M(self):
        if self.P is None:
            return None
        return LOP(self.shape, matvec=self._matvecM)

    @property
    def Jsparse(self):
        if self.K is None:
            return
        J = self.K.copy()  # Should we copy?
        if self.nbcseq:
            c0, c1, c2 = self.bcseqc
            bcseq0 = self.bcseq0
            bcseq1 = self.bcseq1
            # sum up rows at master nodes
            J[bcseq0] += J[bcseq1]
            # add equality constraint at slave nodes
            J[bcseq1] = 0.
            J[bcseq1, bcseq0] = c0/c1
            J[bcseq1, bcseq1] = 1.
        if self.nbcs:
            bcsr = self.bcsr
            J[bcsr] = 0
            J[bcsr, bcsr] = 1.
        if self.nlm:  # bcsr zeroed already in B
            B = self.B
            BT = self.BT
            J = scipy.sparse.bmat([[J, BT], [B, None]])
        return J

    def lop(self, x=None):
        self.update_model(x)  # TODO: is this is wrong becase of x_bcsr?
        return self.l

    def residual(self, x=None):
        self.update_model(x)  # or update rather?
        self.x
        return self.r

    def lsolve(self, **options):
        self.update_model()
        if not hasattr(self, 'K'):  # not initialized yet?
            self.update()
        x0 = self.x
        b = self.b
        J = self.J
        M = self.M
        options = self.getopts(options)
        self.x, info = self.method.method(J, b, x0, M=M, **options)
    solve = lsolve

    def update(self, x=None, f=None):  # f is here for scipy compatibility
        # assign x to model and call its update if it exists
        # self.update_model(x)
        # clear current data grab new data from model
        self.clear()
        self.x0 = self.x  # here we set also all shaping data
        self._update_K()
        self._update_B()
        self._update_P()
        self._update_Klm()
        self._update_Mlm()

    def clear(self):
        attrs = ['nlhs', 'shape_lhs', 'dtype', 'nlm', 'nbcs', 'bcsr', 'x_bcsr',
                 'n', 'shape', 'x0', 'bcseq0', 'bcseq1', 'bcseqc', 'nbcseq',
                 'x_bcseq0', 'x_bcseq1',
                 'K', 'P', 'B', 'BT', 'Klm', 'Mlm',
                 'c', 'K0', 'B0', 'BT0', 'I', 'Ilm', 't0', 'y0', 'f0']
        for a in attrs:
            if hasattr(self, a):
                delattr(self, a)

    def update_model(self, x=None):
        self.x = x    # Should we move x.setter here?
        if hasattr(self.model, 'update'):
            self.model.update()


class DirectSolver(LinearSolver):
    pass


class LUSolver(LinearSolver):
    default_preconditioner = precon.lu


try:
    class UmfpackSolver(DirectSolver):
        default_preconditioner = precon.umfpack
except AttributeError:
    pass


class IterativeSolver(LinearSolver):
    default_preconditioner = precon.inexact
    default_method = methods.iterative


class LgmresSolver(LinearSolver):
    default_preconditioner = precon.inexact
    default_method = methods.lgmres


class GmresSolver(LinearSolver):
    default_preconditioner = precon.inexact
    default_method = methods.gmres


class CgsSolver(LinearSolver):
    default_preconditioner = precon.inexact
    default_method = methods.cgs


class BicgstabSolver(LinearSolver):
    default_preconditioner = precon.inexact
    default_method = methods.bicgstab


try:
    class AmgclCGSolver(IterativeSolver):
        default_preconditioner = precon.amgcl_cg
        # default_method = methods.direct

except AttributeError:
    pass

# if pyamg is present
try:
    #    class AmgBicgstabSolver(LinearSolver):
    #        default_preconditioner = precon.inexact
    #        default_method = methods.amg_bicgstab

    #    class AmgFgmresSolver(LinearSolver):
    #        default_preconditioner = precon.inexact
    #        default_method = methods.amg_fgmres
    #
    #    class AmgGmresSolver(LinearSolver):
    #        default_preconditioner = precon.inexact
    #        default_method = methods.amg_gmres
    pass
except:
    pass


# add utility functions for solving linear systems
def model_from_data(A, b, r=None, x0=None, B=None, c=None):
    class M(object):
        jac = A
        rhs = b
        if r is not None:
            bcs = r
        if x0 is not None:
            lhs = x0.copy()
        if B is not None:
            jaclm = B
        if c is not None:
            lhslm = c
    return M()


def dsolve(A, b, bcs=None, x0=None, B=None, c=None, **options):
    m = model_from_data(A, b, bcs, x0, B, c)
    s = DirectSolver(m, **options)
    s.solve()
    if B is not None:
        return m.lhs, m.lhslm
    return m.lhs


def isolve(A, b, bcs=None, x0=None, B=None, c=None, **options):
    m = model_from_data(A, b, bcs, x0, B, c)
    s = IterativeSolver(m, **options)
    s.solve()
    if B is not None:
        return m.lhs, m.lhslm
    return m.lhs


if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 27 21:30:46 2014

@author: mwojc
"""
from fempy.models.heat import Heat


def test_p2():
    from fempy.geometry import Rectangle
    d = Rectangle(0.3, 0.1, elsize_factor=0.5).gmsh.domain(order=1)

    m = Heat(d)

    m.kij[:] = [[1., 0], [0, 1.]]

    m.alpha['left + bottom'] = 8.
    m.epsilon['left + bottom'] = 0.85
    m.Text['left + bottom'] = 293.15

    m.alpha['right + top'] = 25.
    m.epsilon['right + top'] = 0.85
    m.Text['right + top'] = 1273.15

    m.solve(verbose=True)
    m.preview()
    return m


def test_s1():
    from fempy.geometry import Rectangle
    d = Rectangle(0.3, 0.1, elsize_factor=0.5).gmsh.domain(order=2)
    m = Heat(d)
    m.kij[:] = [[1., 0], [0, 1.]]

    m.alpha['left'] = 8.
    m.epsilon['left'] = 0.85
    m.Text['left'] = 293.15

    m.alpha['right'] = 25.
    m.epsilon['right'] = 0.85
    m.Text['right'] = 1273.15

    m.solve(verbose=True)
    m.preview()
    return m


def test_p2_3D():
    from fempy.geometry import Gmsh
    from pygmsh.built_in import Geometry
    pg = Geometry()
    pg.add_box(0, 0.3, 0, 0.1, 0, 0.1, 0.0125)
    g = Gmsh(pg, entities={2: {19: 'right',
                               20: 'top',
                               21: 'front',
                               22: 'back',
                               23: 'bottom',
                               24: 'left'}})
    d = g.domain(order=1)
    m = Heat(d)
    m.kij[:] = [[1., 0, 0], [0, 1., 0], [0, 0, 1.]]

    m.alpha['left+bottom'] = 8.
    m.epsilon['left+bottom'] = 0.85
    m.Text['left+bottom'] = 293.15

    m.alpha['right+top'] = 25.
    m.epsilon['right+top'] = 0.85
    m.Text['right+top'] = 1273.15

    m.solve(verbose=True)
    m.preview()
    return m


def test_s1_3D():
    from fempy.geometry import Gmsh
    from pygmsh.built_in import Geometry
    pg = Geometry()
    pg.add_box(0, 0.3, 0, 0.1, 0, 0.1, 0.0125)
    g = Gmsh(pg, entities={2: {19: 'right',
                               20: 'top',
                               21: 'front',
                               22: 'back',
                               23: 'bottom',
                               24: 'left'}})
    d = g.domain(order=1)
    m = Heat(d)
    m.kij[:] = [[1., 0, 0], [0, 1., 0], [0, 0, 1.]]

    m.alpha['left'] = 8.
    m.epsilon['left'] = 0.85
    m.Text['left'] = 293.15

    m.alpha['right'] = 25.
    m.epsilon['right'] = 0.85
    m.Text['right'] = 1273.15

    m.solve(verbose=True)
    m.preview()
    return m


def test_3D():
    from fempy.geometry import Gmsh
    from pygmsh.built_in import Geometry
    pg = Geometry()
    pg.add_box(0, 0.3, 0, 0.1, 0, 0.1, 0.0125)
    g = Gmsh(pg, entities={2: {19: 'right',
                               20: 'top',
                               21: 'front',
                               22: 'back',
                               23: 'bottom',
                               24: 'left'}})
    d = g.domain(order=1)
    m = Heat(d)
    m.kij[:] = [[1., 0, 0], [0, 1., 0], [0, 0, 1.]]

    m.alpha['left + bottom + front'] = 8.
    m.epsilon['left + bottom + front'] = 0.85
    m.Text['left + bottom + front'] = 293.15

    m.alpha['right + top + back'] = 25.
    m.epsilon['right + top + back'] = 0.85
    m.Text['right + top + back'] = 1273.15

    m.solve(verbose=True)
    m.preview()
    return m


def test_1D():
    from fempy.geometry import Point, Line
    A = Point([0, 0, 0], groups='left', elsize=0.0125)
    B = Point([0.3, 0, 0], groups='right', elsize=0.0125)
    AB = Line(A, B, groups='line')
    d = AB.gmsh.domain(2)

    m = Heat(d)
    m.kij[:] = [[1.]]

    m.alpha['left'] = 8.
    m.epsilon['left'] = 0.85
    m.Text['left'] = 293.15

    m.alpha['right'] = 25.
    m.epsilon['right'] = 0.85
    m.Text['right'] = 1273.15

    m.solve(verbose=True)
    m.preview()
    return m


def test_p2_time():
    from fempy.solvers import ivp
    from fempy.fields.operators import ongauss
    from fempy.geometry import Rectangle
    import numpy as np

    # Class
    class MyHeat(Heat):
        default_solver = ivp.BDFSolver

        def setup(self):
            m = self
            m.alpha['left + bottom'] = 8.
            m.epsilon['left + bottom'] = 0.85
            m.alpha['right + top'] = 25.
            m.epsilon['right + top'] = 0.85
            m.kij[:] = [[1, 0], [0, 1.]]
            m.rho[:] = 2300.
            m.cp[:] = 800.
            m.T[:] = 293.15
            m.Text[:] = 293.15
            m.time0 = 0.
            m.time1 = 120*60.
            self.bcs['left'] = True
            #self.bcseq = [['bottom_right', 'top_right']]

        def kij_update(self):
            T = ongauss(self.T)
            kij = self.kij
            T0 = 293.15
            k0 = 1.4
            T1 = 1273.15
            k1 = 1.0
            a = (k1 - k0) / (T1 - T0)
            I = np.eye(self.domain.ndime)

            kij[..., 0, 0] = k0 + a * (T - T0)
            kij[..., 1, 1] = k0 + a * (T - T0)
            kij[T <= T0] = I * k0
            kij[T >= T1] = I * k1

        def Text_update(self):
            t = self.time
            t0 = 0.
            Text0 = 293.15
            t1 = 60*120.
            Text1 = 1273.15
            a = (Text1 - Text0) / (t1 - t0)
            self.Text['left + bottom'] = Text0
            self.Text['right + top'] = Text0 + a*(t - t0)
            self.T['left'] = Text0 + t/9.

    # Initial state
    d = Rectangle(0.3, 0.1, elsize_factor=0.5).gmsh.domain(order=2)
    m = MyHeat(d)
    m.solve(verbose=True)
    m.preview()
    return m


test_s1()
test_p2()
test_s1_3D()
test_p2_3D()
test_3D()
test_1D()
m = test_p2_time()



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Nov  9 12:18:04 2018

@author: mwojc
"""
from __future__ import absolute_import, division, print_function,\
                       unicode_literals
import sys
version = sys.version_info
python = version[0]


if python == 2:
    # Builtins
    import __builtin__ as builtins

    # Pickling
    import cPickle as picklemod

    class PickleCompat(object):
        def dump(self, obj, file):
            return picklemod.dump(obj, file, protocol=2)

        def load(self, file):
            return picklemod.load(file)
    pickle = PickleCompat()

    # Reloading
    reload = builtins.reload


if python == 3:
    # Builtins
    import builtins

    # Pickling
    import pickle as picklemod

    class PickleCompat(object):
        def dump(self, obj, file):
            return picklemod.dump(obj, file, protocol=2, fix_imports=True)

        def load(self, file):
            try:
                return picklemod.load(file)
            except UnicodeDecodeError:
                file.seek(0)
                return picklemod.load(file, encoding='latin1')
    pickle = PickleCompat()

    # Reloading
    from importlib import reload



# -*- coding: utf-8 -*-
"""
Created on Tue Jun 28 12:54:09 2016

@author: mwojc
"""
from __future__ import print_function
import pytest
import inspect
import numpy as np
import scipy.sparse as sparse
import scipy.sparse.linalg
import fempy.solvers.methods as methods
import fempy.solvers.precon as precon

# List of linear solver methods
methods = [ls for ls in methods.__dict__.values()
           if inspect.isclass(ls) and issubclass(ls, methods.linear)]


class Test_Linear_Methods:
    def test_linear_methods(self, capsys):
        n = 20
        tol = 1e-10
        Ad = np.random.rand(n, n)
#        Ad = Ad.dot(Ad.T)       # positively definite
#        Ad /= Ad.max()
#        Ad = (Ad + Ad.T)/2.     # symmetric
        As = sparse.csr_matrix(Ad)
        b = np.random.rand(n)
        x0 = np.linalg.solve(Ad, b)
        with capsys.disabled():
            print("\n")
            for ls in methods:
                print("Testing {} method".format(ls.__name__))
                ls = ls()
                if not ls.direct:  # iterative
                    x, _ = ls.solve(Ad, b, M=precon.ilu()(Ad), tol=tol)
                    assert np.allclose(x, x0)
                    x, _ = ls.solve(As, b, M=precon.lu()(Ad), tol=tol)
                    assert np.allclose(x, x0)
                    x, _ = ls.solve(As, b, tol=tol)
                    assert np.allclose(x, x0)
#                    x, _ = ls.solve(As, b, M=precon.AmgSas()(As), tol=tol)
#                    assert np.allclose(x, self.x)
                else:  # direct
                    with pytest.raises(ValueError):
                        x, _ = ls.solve(Ad, b)
                    x, _ = ls.solve(Ad, b, M=precon.lu()(Ad))
                    assert np.allclose(x, x0)
                    x, _ = ls.solve(As, b, M=precon.lu()(As))
                    assert np.allclose(x, x0)

if __name__ == '__main__':
    # import pytest
    pytest.main([str(__file__), '-v'])  # Run tests from current file



# -*- coding: utf-8 -*-
"""
Created on Sun Nov 16 13:49:09 2014

@author: mwojc
"""
import numpy as np
from fempy.fields.fields import Field
from fempy.fields.nfields import nField


class gField(Field):
    def __new__(cls, domain, data):
        if isinstance(data, gField):
            if domain is data.domain:
                return data
        if isinstance(data, nField):
            if domain is data.domain:
                from fempy.fields.operators import ongauss
                return ongauss(data)
        ne = domain.nelem
        ng = domain.etype.ngauss
        nd = domain.ndime
        data = np.asarray(data)
        prefix = 2
        if not data.shape[:prefix] == (ne, ng):
            default = data
            data = np.empty((ne, ng) + data.shape, dtype=data.dtype)
            data[:] = default
        # Field factory
        sshape = data.shape[prefix:]
        if sshape == ():
            cls = gScalar
        elif len(sshape) == 1 and sshape == (nd,):
            cls = gVector
        elif len(sshape) == 1 and sshape != (nd,):
            cls = gList
        elif len(sshape) == 2 and sshape == (nd,)*2:
            cls = gTensor
        elif len(sshape) == 2 and sshape != (nd,)*2:
            cls = gMatrix
        elif sshape == (nd,)*3:
            cls = gTensor3
        elif sshape == (nd,)*4:
            cls = gTensor4
        else:
            pass
        #
        f = Field.__new__(cls, data, prefix=prefix)
        f.domain = domain
        return f

    def __array_finalize__(self, field):
        if field is None:
            return
        Field.__array_finalize__(self, field)
        self.domain = getattr(field, 'domain', None)

    @property
    def groups(self):
        return self.domain.egroups

    def factory(self, data):
        return gField(self.domain, data)


class gScalar(gField):
    def __new__(cls, domain, data=None):
        ne = domain.nelem
        ng = domain.etype.ngauss
        if data is None:
            data = np.zeros((ne, ng))
        data = np.asarray(data)
        assert data.shape == (ne, ng) or data.shape == ()
        return gField.__new__(cls, domain, data)


class gList(gField):
    def __new__(cls, domain, data):
        ne = domain.nelem
        ng = domain.etype.ngauss
        if isinstance(data, int):
            data = np.zeros((ne, ng, data))
        else:
            data = np.asarray(data)
            assert len(data.shape) in (1, 3)
            if len(data.shape) == 3:
                assert data.shape[:2] == (ne, ng)
        return gField.__new__(cls, domain, data)


class gVector(gList):
    def __new__(cls, domain, data=None):
        ne = domain.nelem
        ng = domain.etype.ngauss
        nd = domain.ndime
        if data is None:
            data = np.zeros((ne, ng, nd))
        data = np.asarray(data)
        if data.shape == ():
            data = np.tile(data, nd)
        assert data.shape == (ne, ng, nd) or data.shape == (nd,)
        return gField.__new__(cls, domain, data)


class gMatrix(gField):
    def __new__(cls, domain, data):
        ne = domain.nelem
        ng = domain.etype.ngauss
        if isinstance(data, int):
            data = np.zeros((ne, ng, data, data))
        else:
            data = np.asarray(data)
            assert len(data.shape) in (1, 2, 4)
            if len(data.shape) == 1:
                assert data.dtype == int and data.shape[0] == 2
                data = np.zeros(tuple(data))
            if len(data.shape) == 4:
                assert data.shape[:2] == (ne, ng)
        return gField.__new__(cls, domain, data)


class gTensor(gMatrix):
    def __new__(cls, domain, data=None):
        ne = domain.nelem
        ng = domain.etype.ngauss
        nd = domain.ndime
        if data is None:
            data = np.zeros((ne, ng, nd, nd))
        data = np.asarray(data)
        if data.shape == ():
            data = np.tile(data, (nd, nd))
        assert data.shape == (ne, ng, nd, nd) or data.shape == (nd,)*2
        return gField.__new__(cls, domain, data)


class gTensor3(gField):
    def __new__(cls, domain, data=None):
        ne = domain.nelem
        ng = domain.etype.ngauss
        nd = domain.ndime
        if data is None:
            data = np.zeros((ne, ng, nd, nd, nd))
        data = np.asarray(data)
        if data.shape == ():
            data = np.tile(data, (nd, nd, nd))
        assert data.shape == (ne, ng, nd, nd, nd) or data.shape == (nd,)*3
        return gField.__new__(cls, domain, data)


class gTensor4(gField):
    def __new__(cls, domain, data=None):
        ne = domain.nelem
        ng = domain.etype.ngauss
        nd = domain.ndime
        if data is None:
            data = np.zeros((ne, ng, nd, nd, nd, nd))
        data = np.asarray(data)
        if data.shape == ():
            data = np.tile(data, (nd, nd, nd, nd))
        assert data.shape == (ne, ng, nd, nd, nd, nd) or data.shape == (nd,)*4
        return gField.__new__(cls, domain, data)


if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')



# -*- coding: utf-8 -*-
"""
Created on Tue Nov 18 23:24:04 2014

@author: mwojc
"""

import numpy as np
from scipy import sparse
from fempy.fields.algebra import dot, transpose
from fempy.fields.fields import Field
from fempy.fields.nfields import nField, nScalar, nVector, nTensor, nTensor3, \
                                 nTensor4
from fempy.fields.gfields import gField, gScalar, gVector, gTensor, gTensor3, \
                                 gTensor4


class BaseOperator(Field):
    def __new__(cls, nfield, operator):
        assert isinstance(nfield, nField), "nField must be provided"
        f = nfield
        d = f.domain
        ne = d.nelem
        ng = d.etype.ngauss
        nn = d.etype.nnode
        prefix = len(f.shape) + 2  # +2 for gn
        assert operator.shape[:prefix] == (ne, ng, nn) + f.sshape
        operator = Field.__new__(cls, operator, prefix=prefix)
        operator.nfield = nfield
        return operator

    def __array_finalize__(self, op):
        if op is None:
            return
        Field.__array_finalize__(self, op)
        self.nfield = getattr(op, 'nfield', None)

    @property
    def groups(self):
        return self.field.domain.egroups

    @property
    def gfield(self):
        fs = 'abcdefgh'[:self.nfield.sufix]  # sshape of underlaying field
        os = 'pqrstuvw'[:self.sufix]  # sshape of operator (and values)
        vals = np.einsum('egn%s%s,en%s->eg%s' % (fs, os, fs, os),
                         self, onelems(self.nfield))
        return gField(self.nfield.domain, vals)

    def factory(self, data=None):
        if data is None:
            data = np.zeros(self.shape)
        op = self.copy()
        op[:] = data
        return op


class Operator(BaseOperator):
    def __new__(cls, field):
        f = field
        egn = f.domain.egn
        if f.sufix > 0:
            ss = f.sshape
            pss = np.prod(ss, dtype=int)
            op = np.zeros((pss, pss) + egn.shape)
            op[range(pss), range(pss)] = egn
            ax = list(range(len(op.shape)))
            op = op.transpose(ax[2:] + [0, 1])
            op = op.reshape(op.shape[:-2] + ss*2)
            egn = op
            # ss = f.sshape
            # eye = np.eye(np.prod(ss)).reshape(ss*2)
            # egn = np.einsum('egn...,...->egn...', egn, eye)
            # TODO: use numpy as_strided here for memory saving
        egn = BaseOperator.__new__(cls, field, egn)
        return egn
operator = Operator


class Nabla(BaseOperator):
    def __new__(cls, field):
        f = field
        egnk = f.domain.egnk
        if f.sufix > 0:
            ss = f.sshape
            pss = np.prod(ss, dtype=int)
            op = np.zeros((pss, pss) + egnk.shape)
            op[range(pss), range(pss)] = egnk
            ax = list(range(len(op.shape)))
            op = op.transpose(ax[2:-1] + [0, 1, -1])
            op = op.reshape(op.shape[:-3] + ss*2 + (op.shape[-1],))
            egnk = op
            # ss = f.sshape
            # eye = np.eye(np.prod(ss, dtype=int)).reshape(ss*2)
            # egnk = np.einsum('egnk...,...->egn...k', egnk, eye)
            # TODO: use numpy as_strided here for memory saving
        egnk = BaseOperator.__new__(cls, field, egnk)
        return egnk
nabla = Nabla


class SmallStrain(BaseOperator):
    def __new__(cls, field):
        assert isinstance(field, nVector), \
            "SmallStrain can only be applied to nVector"
        nbl = Nabla(field)
        sstr = 0.5*(nbl + transpose(nbl))
        sstr = sstr.view(cls)
        # egnk = BaseOperator.__new__(cls, field, egnk)
        return sstr
small_strain = SmallStrain


class SmallRotation(BaseOperator):
    def __new__(cls, field):
        assert isinstance(field, nVector), \
            "SmallRotation can only be applied to nVector"
        nbl = Nabla(field)
        sstr = 0.5*(nbl - transpose(nbl))
        sstr = sstr.view(cls)
        # egnk = BaseOperator.__new__(cls, field, egnk)
        return sstr
small_rotation = SmallRotation


class Divergence(BaseOperator):
    def __new__(cls, field):
        assert isinstance(field, (nVector, nTensor, nTensor3, nTensor4) +
                                 (gVector, gTensor, gTensor3, gTensor4)), \
            "Divergence can be computed for vectors and tensors only"
        nbl = Nabla(field)
        # dv = div(nbl)
        dv = np.einsum('egn...ii->egn...', nbl)  # just like in div
        dv = BaseOperator(field, dv)
        dv = dv.view(cls)
        return dv
divergence = Divergence


def onnodes(f, order=0, power=0):
    if isinstance(f, nField):
        return f
    # Get weights
    d = f.domain
    ec = d.ecoors
    gc = d.gcoors
    egni = np.einsum('eni,g->egni', ec, np.ones(d.etype.ngauss))
    egni = (egni.transpose(2, 0, 1, 3) - gc).transpose(1, 2, 0, 3)  # egni
    egn = np.linalg.norm(egni, axis=-1)  # egn, distance
    egn = 1./egn**power
    e1n = egn.sum(1, keepdims=True)
    egn = egn/e1n  # normalized weights
    # Calculate nodal values using weights
    en = np.einsum('eg...,egn...->en...', f, egn)
    n = nField(d, np.zeros(f.sshape))
    np.add.at(n, d.conec, en)  # assemble
    # And average them (not weighted here)
    n = dot(n, 1./d.ndeg)
    nres = n
    # if order > 1 calculate derivatives and update
    for i in range(order):
        egi = grad(nres)  # interpolated field derivatives (with shape func.)
        df = np.einsum('eg...i,egni->egn...', egi, egni)
        df = np.einsum('egn...,egn...->en...', df, egn)
        dn = nField(d, np.zeros(f.sshape))
        np.add.at(dn, d.conec, df)  # assemble
        dn = dot(dn, 1./d.ndeg)
        nres = n + dn
    return nres


def onelems(f):
    return onnodes(f)[f.domain.conec]


def ongauss(f):
    if isinstance(f, gField):
        eg = f
    else:
        # gn = domain.etype.gshp
        egn = f.domain.egn
        eg = np.einsum('egn...,en...->eg...', egn, onelems(f))
        eg = gField(f.domain, eg)
    return eg


def onboundary(f):
    assert isinstance(f, nField), "nField must be provided"
    b = f.domain.boundary
    return nField(b, f['boundary'])


def grad(f):
    egnk = f.domain.egnk
    egk = np.einsum('egnk...,en...->eg...k', egnk, onelems(f))
    return gField(f.domain, egk)


def div(f):
    # assert f.sufix and f.shape[-1] == f.domain.ndime, \
    #     "Last field axis must be equal to the domain dimension."
    assert isinstance(f, (nVector, nTensor, nTensor3, nTensor4) +
                         (gVector, gTensor, gTensor3, gTensor4)), \
        "Divergence can be computed for vectors and tensors only"
    g = grad(f)
    d = np.einsum('eg...ii->eg...', g)
    return gField(f.domain, d)


def integrate(f):
    # TODO: This function should be removed or renamed because its name is
    #       misleading. It actually assembles divergence of the field,
    #       equivalent to:
    #       lform(f, v=nabla(f))
    d = f.domain
    eg = dot(ongauss(f), d.gvol)  # to gauss points and multiply by volumes
    en = np.einsum('eg...k,egn...k->en...', eg, d.egnk)
    F = nField(d, np.zeros(f.sshape[:-1]))
    np.add.at(F, d.conec, en)  # assemble
#    if isinstance(f, gField):
#        print 'aaa'
#        fn = nField(f.domain, np.zeros(f.sshape))
#        print fn.shape
#    else:
#        fn = f
#    F = lform(f, nabla(fn))
    return F


def integrate2(f):
    f = ongauss(f)
    v = f.domain.gvol
    return np.einsum('eg...,eg->...', f, v)


def lform(f, v=None):
    # R_enA = F_egBA * N_egnAB
    # A = i_1 ... i_n
    # B = j_1 ... j_n
    d = f.domain
    # Use 'operator' by default
    if v is None:
        fn = f
        if isinstance(f, gField):
            fn = nField(f.domain, np.zeros(f.sshape))
        v = operator(fn)
    eg = dot(ongauss(f), d.gvol)  # to gauss points and multiply by volumes
    # TODO: checks for f, v shapes
    j = 'rstuvwxy'[:eg.sufix]
    subscripts = 'eg%s...,egn...%s->en...' % (j, j)
    en = np.einsum(subscripts, eg, v)  # integrate to nodes, to be assembled
    # Assemble
    F = np.zeros((d.eidx.nidx.g.na,) + en.shape[2:])
    np.add.at(F, d.eidx.conec_a, en)
    return F  # TODO: we should return nfield here


def lformu(f):
    d = f.domain
    gf = ongauss(f)
    gvals = np.einsum('eg...,eg->eg...', gf, d.gvol)
    evals = gvals.sum(1)
    envals = np.zeros((d.nelem, d.etype.nnode)+f.sshape)
    envals[:] = evals[:, None]/d.etype.nnode
    F = np.zeros((d.eidx.nidx.g.na,) + f.sshape)
    np.add.at(F, d.eidx.conec_a, envals)
    return F


def eij(u, v):
    as_strided = np.lib.stride_tricks.as_strided
    eipos1 = u.nfield.eipos
    m1, n1 = eipos1.shape
    sz = eipos1.itemsize
    eipos2 = v.nfield.eipos
    m2, n2 = eipos2.shape
    sz = eipos2.itemsize
    ei = as_strided(eipos1, shape=(m1, n1, n2), strides=(n1*sz, sz, 0))
    ej = as_strided(eipos2, shape=(m2, n1, n2), strides=(n2*sz, 0, sz))
    return ei, ej


def qform(A, u=None, v=None):
    d = A.domain
    # Use 'nabla' by default
    if u is None:
        An = A
        if isinstance(A, gField):
            An = nField(A.domain, np.zeros(A.sshape[A.sufix//2+1:]))
        u = nabla(An)
    if v is None:
        v = u
    if not isinstance(u, BaseOperator):
        u = BaseOperator(v.nfield, u)
    if not isinstance(v, BaseOperator):
        v = BaseOperator(u.nfield, v)
    # assert u.nfield is v.nfield
    # We are ready to integrate and assemble
    A = dot(ongauss(A), d.gvol)  # to gauss points and multiply by volumes
    # TODO: checks for A, u, v shapes
    i = 'abcd'[:u.nfield.sufix]
    j = 'fhyz'[:v.nfield.sufix]
    pq = 'ijklopq'[:u.sufix]
    rs = 'rstuvwx'[:v.sufix]
#    k = np.einsum('eg%s%s, egm%s%s, egn%s%s -> em%sn%s'
#                  % (pq, rs, i, pq, j, rs, i, j), A, u, v)  # SLOW!!!
    A = np.einsum('egm%s%s, eg%s%s -> egm%s%s' % (i, pq, pq, rs, i, rs), u, A)
    k = np.einsum('egm%s%s, egn%s%s -> em%sn%s' % (i, rs, j, rs, i, j), A, v)
    i, j = eij(u, v)  # u.nfield.eij
    shape = (u.nfield.nsize, v.nfield.nsize)
    k = k.ravel()
    i = i.ravel()
    j = j.ravel()
    K = sparse.coo_matrix((k, (i, j)), shape=shape).tocsc()  # assemble
    return K


def dirichlet(f):
    return f.__class__(f.domain, False)


if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 27 21:30:46 2014

@author: mwojc
"""
from fempy.geometry.shapes import Rectangle
from fempy.models.elastic import PlaneStrain

d = Rectangle(elsize=0.05).gmsh.domain(order=2)
m = PlaneStrain(d)

# material properties
m.ev(20000, 0.3)
m.alpha(1e-4)
m.rho[:] = 2.

# loads
m.p['right'] = -10.
m.qi['left'] = [10, 0.]

# temperature increase
m.T += 3.

# dirichlet bcs
m.u['top', 1] = -0.0005
m.bcs['top', 1] = True
m.bcs['bottom'] = True

m.solve()
m.preview('u')
# m.preview('all')


# FempyView is depreciated
## View
#from fempy.view import FempyView
##d.coors += m.u*500  # add displacements to mesh
#view = FempyView(d, m.u*500, mesh_visible=True)
#view.cell_scalars(sigma[..., 0, 1, 1], name='Syy', legend=True)
##from fempy.fields.operators import onnodes
##view.point_scalars(onnodes(sigma)[..., 1, 1], name='Syy', legend=True)



# -*- coding: utf-8 -*-
"""
Created on Fri Oct 13 09:20:09 2017

@author: mwojc
"""
from __future__ import print_function
from fempy.compat import reload
import inspect
import numpy as np
from fempy.solvers import linear
from fempy.solvers import nonlin
from fempy.solvers import precon
from fempy.solvers import methods


def _get_solvers():
    reload(methods)
    reload(precon)
    reload(linear)
    reload(nonlin)
    mc = [ls for ls in methods.__dict__.values()
          if inspect.isclass(ls) and issubclass(ls, methods.linear)]

    pc = [ls for ls in precon.__dict__.values()
          if inspect.isclass(ls) and issubclass(ls, precon.preconditioner)]

    lsc = [ls for ls in linear.__dict__.values()
           if inspect.isclass(ls) and issubclass(ls, linear.LinearSolver)]

    nsc = [ls for ls in nonlin.__dict__.values()
           if inspect.isclass(ls) and issubclass(ls, nonlin.NonLinearSolver)]
    return mc, pc, lsc, nsc


def _solver_test(solver, model, n=1, x0=None, **opts):
    import time
    T = 0.
    err = 0
    lhs_init = model.lhs.copy()
    if hasattr(model, 'lhslm'):
        lhslm_init = model.lhslm.copy()
    for i in range(n):
        m = model
        m.lhs[:] = lhs_init
        if hasattr(model, 'jaclm'):
            m.lhslm[:] = lhslm_init
        ls = solver(m)  # new solver is created

        if ls.method.direct and m.jac is None:
            T = 10000000000.
            break

        if i == 0:
            # Print message
            mname = m.__class__.__name__
            nname = ls.__class__.__name__
            lname = ls.method.__class__.__name__
            pname = ls.preconditioner.__class__.__name__ \
                if ls.preconditioner is not None else 'none'
            print("Testing %s model against %s with %s method and"
                  " %s preconditioner" % (mname, nname, lname, pname))
            import sys
            sys.stdout.flush()

        t0 = time.time()
        ls.solve(**opts)
        T += time.time() - t0

        # Test solution
        x = ls.x
        # x0 = m.x0 if hasattr(m, 'x0') else x0
        if x0 is not None:
            if not np.allclose(x, x0.ravel()):
                print("    Solver did not reach the reference solution...")
                print("    Average x error: %s" % np.mean(abs(x - x0.ravel())))
                err = 1
    m.lhs[:] = lhs_init
    if hasattr(model, 'lhslm'):
        m.lhslm[:] = lhslm_init
    T = T/float(n)
    print("    Time spent in solver: %s\n" % T)
    return T, err


def model_test(model, solver=None, method=None, preconditioner=None,
               n=1, x0=None, strict=False, **options):
    mc, pc, lsc, nsc = _get_solvers()
    # Choose solvers
    if solver is None or len(solver) == 0:
        solver = [model.default_solver]
    if solver == 'all':
        solver = lsc + nsc
    if solver == 'linear':
        solver = lsc
    if solver == 'nonlin':
        solver = nsc
    # Choose methods
    if method is None or len(method) == 0:
        method = [model.solver.default_method]
    if method == 'all':
        method = mc
    if method == 'direct':
        method = [m for m in mc if m.direct]
    if method == 'iterative':
        method = [m for m in mc if not m.direct]
    # Choose preconditioners
    if preconditioner is None or len(preconditioner) == 0:
        preconditioner = [model.solver.default_preconditioner]
    if preconditioner == 'all':
        preconditioner = pc + [None]
    if preconditioner == 'exact':
        preconditioner = [p for p in pc if p.exact]
    if preconditioner == 'inexact':
        preconditioner = [p for p in pc if not p.exact]

    name = model.__class__.__name__
    if hasattr(model, 'info'):
        print(model.info)
    Tl = 100000000.
    bs = ''
    bm = ''
    bp = ''
    for sol in solver:
        sname = sol.__name__
        if hasattr(model, 'blacklist') and sname in model.blacklist:
            continue
        for mth in method:
            mname = mth.__name__
            sol.default_method = mth
            for p in preconditioner:
                pname = p.__name__ if p is not None else 'none'
                # skip direct methods when preconditioner is inexact
                if sol.default_method.direct:
                    if p is None or not p.exact:
                        continue
                if hasattr(model, 'pblacklist') and \
                        mname in model.pblacklist and \
                        (len(model.pblacklist[mname]) == 0 or
                         pname in model.pblacklist[mname]):
                    continue
                sol.default_preconditioner = p
                # print(name, sname, mname, pname)
                tl, err = _solver_test(sol, model, n=n, x0=x0, **options)
                if tl < Tl and not err:
                    Tl = tl
                    bs = sname
                    bm = mname
                    bp = pname
                if err and strict:
                    raise Exception('Solutions do not match')
    print('Best time for model %s: %s s' % (name, Tl))
    print('Best solver: %s with %s method and %s preconditioner'
          % (bs, bm, bp))


if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')



# -*- coding: utf-8 -*-
"""
Created on Wed Nov 19 10:45:27 2014

@author: mwojc
"""

##########################################################
import numpy as np
from fempy.fields.fields import Field
from fempy.fields.algebra import transpose


class Test_Field:
    r0 = np.random.rand(10)
    r1 = np.random.rand(10, 3)
    r2 = np.random.rand(10, 2, 2)
    r3 = np.random.rand(10, 3, 3, 3)
    r4 = np.random.rand(10, 4, 4, 4, 6)

    def test_getitem(self):
        f = Field(self.r4, groups={'a': [1, 2, 3], 'b': [5, 4]})
        assert np.allclose(f['a'], self.r4[[1, 2, 3]])
        assert np.allclose(f['a', 0], self.r4[[1, 2, 3], 0])
        assert np.allclose(f['b'][0], self.r4[5])

        # Groups work only on first axis
        f = Field(self.r4, prefix=2, groups={'a': [1, 2, 3], 'b': [5, 4]})
        assert np.allclose(f['a'], self.r4[[1, 2, 3]])
        assert np.allclose(f['a', 0], self.r4[[1, 2, 3], 0])
        assert np.allclose(f['b'][0], self.r4[5])

    def test_setitem(self):
        f = Field(self.r2, groups={'a': [1, 2, 3], 'b': [5, 4]})
        f['a'] = 3.
        assert np.allclose(f['a'], 3.)
        f['b'] = [[1, 2], [3, 4]]
        assert np.allclose(f['b', 1, 1], 4.)
        assert np.allclose(f['b'][0], [[1, 2], [3, 4]])
        f['b'] = [4, 5]
        assert np.allclose(f['b'][0], [[4, 5], [4, 5]])
        f['b', 0] = 7.
        assert np.allclose(f['b'][0], [[7, 7], [4, 5]])
        f['b', 1, 0] = 17.
        assert np.allclose(f['b'][0], [[7, 7], [17, 5]])

        # Groups work only on first axis
        f = Field(self.r2, prefix=2, groups={'a': [1, 2, 3], 'b': [5, 4]})
        f['b'] = [[1, 2], [3, 4]]
        assert np.allclose(f['b', 1, 1], 4.)
        assert np.allclose(f['b'][0], [[1, 2], [3, 4]])
        f['b'] = [4, 5]
        assert np.allclose(f['b'][0], [[4, 5], [4, 5]])

    def test_scalar_field_creation(self):
        f = Field(self.r0)
        assert f.base is self.r0
        assert f.pshape == (10,)
        assert f.paxes == (0,)
        assert f.sshape == ()
        assert f.saxes == ()
        assert (f == transpose(f)).all()
        f = Field(self.r1, 2)
        assert f.base is self.r1
        assert f.pshape == (10, 3)
        assert f.paxes == (0, 1)
        assert f.sshape == ()
        assert f.saxes == ()
        assert (f == transpose(f)).all()

    def test_vector_field_creation(self):
        f = Field(self.r1)
        assert f.base is self.r1
        assert f.pshape == (10,)
        assert f.paxes == (0,)
        assert f.sshape == (3,)
        assert f.saxes == (1,)
        assert (f == transpose(f)).all()
        f = Field(self.r2, 2)
        assert f.base is self.r2
        assert f.pshape == (10, 2)
        assert f.paxes == (0, 1)
        assert f.sshape == (2,)
        assert f.saxes == (2,)
        assert (f == transpose(f)).all()

    def test_tensor_field_creation(self):
        f = Field(self.r2)
        assert f.base is self.r2
        assert f.pshape == (10,)
        assert f.paxes == (0,)
        assert f.sshape == (2, 2)
        assert f.saxes == (1, 2)
        fT = np.transpose(self.r2, (0, 2, 1))
        assert (fT == transpose(f)).all()
        f = Field(self.r3, 2)
        assert f.base is self.r3
        assert f.pshape == (10, 3)
        assert f.paxes == (0, 1)
        assert f.sshape == (3, 3)
        assert f.saxes == (2, 3)
        fT = np.transpose(self.r3, (0, 1, 3, 2))
        assert (fT == transpose(f)).all()

    def test_any_field_creation(self):
        f = Field(self.r4, 2)
        assert f.base is self.r4
        assert f.pshape == (10, 4)
        assert f.paxes == (0, 1)
        assert f.sshape == (4, 4, 6)
        assert f.saxes == (2, 3, 4)
        fT = np.transpose(self.r4, (0, 1, 4, 3, 2))
        assert (fT == transpose(f)).all()

    def test_ufuncs(self):
        # ufuncs should return field only if pshape of the result is the same
        f = Field(self.r0)  # scalar
        assert isinstance(np.sum(f), float)
        f = Field(self.r2)  # tensor
        assert isinstance(np.sum(f, 2), Field)
        assert isinstance(np.sum(f, 0), np.ndarray)


if __name__ == '__main__':
    import pytest
    pytest.main([str(__file__), '-v'])  # Run tests from current file



# -*- coding: utf-8 -*-
"""
Created on Sat Nov 22 22:47:46 2014

@author: mwojc
"""

import numpy as np
import fempy.domain.elems as elems
from fempy.domain.domain import Domain


class TestElements(object):
    def test_attributes(self):
        for k, v in elems.__dict__.items():
            if isinstance(v, type) and issubclass(v, elems.Element) and \
                    v is not elems.Element:
                e = v()
                e.ncoors
                e.gcoors
                e.gweights
                e.shp
                e.shpd
                e.shpd2
                e.gshp
                e.gshpd
                e.gshpd2
                e.ngauss
                e.nnode
                e.ndime

                e.faces
                e.btype
                e.bncoors
                e.bgcoors
                e.bgshp
                e.bgshpd
                e.bgshpd2
                e.nfaces

    def test_line21(self):
        allc = np.allclose
        e = elems.Line21()
        assert allc(e.nshp, np.eye(e.nnode))

        d = Domain(e.ncoors, np.arange(e.nnode)[None, :])  # single element d
        b = d.boundary
        assert allc(d.vol, 2.0)
        assert allc(b.vol, 2.0)

    def test_tetr41_and_tetr104(self):
        allc = np.allclose
        for e in elems.Tetr41(), elems.Tetr104():
            assert allc(e.nshp, np.eye(e.nnode))

            # single element domain
            d = Domain(e.ncoors, np.arange(e.nnode)[None, :])
            assert allc(d.vol, 0.1666666666667)

            # Faces
            faces = d.boundary
            assert allc(faces.evol, [0.5, 0.8660254, 0.5, 0.5])
            assert allc(faces.vol, 2.36602540378)
            n = faces.normal
            assert allc(n[0], [0, -1, 0])
            assert allc(n[1], [0.57735027, 0.57735027, 0.57735027])
            assert allc(n[2], [-1, 0, 0])
            assert allc(n[3], [0, 0, -1])

            # Edges
            n = faces[0].boundary.normal
            assert allc(n[0], [0, 0, -1])
            assert allc(n[1], [0.70710678, 0., 0.70710678])
            assert allc(n[2], [-1, 0., 0.])
            assert allc(faces[0].boundary.evol, [1., 1.41421356, 1.])

            n = faces[1].boundary.normal
            assert allc(n[0], [0.40824829, 0.40824829, -0.81649658])
            assert allc(n[1], [-0.81649658, 0.40824829, 0.40824829])
            assert allc(n[2], [0.40824829, -0.81649658, 0.40824829])
            assert allc(faces[1].boundary.evol,
                        [1.41421356, 1.41421356, 1.41421356])

            n = faces[2].boundary.normal
            assert allc(n[0], [0, 0, -1])
            assert allc(n[1], [0,  -1,  0.])
            assert allc(n[2], [0., 0.70710678, 0.70710678])
            assert allc(faces[2].boundary.evol, [1., 1., 1.41421356])

            n = faces[3].boundary.normal
            assert allc(n[0], [-1, 0, 0])
            assert allc(n[1], [0.70710678, 0.70710678, 0.])
            assert allc(n[2], [0, -1, 0.])
            assert allc(faces[3].boundary.evol, [1., 1.41421356, 1.])


if __name__ == '__main__':
    import pytest
    pytest.main([str(__file__), '-v'])  # Run tests from current file



# -*- coding: utf-8 -*-
"""
Created on Fri Feb 21 18:08:22 2014

@author: mwojc
"""
import numpy as np


def norm(vec):
    return np.sqrt(np.einsum('...i,...i', vec, vec))


def coors3D(coors):
    if isinstance(coors, (int, float)):
        coors = [coors]
    coors = np.array(coors, dtype=float)
    shape = coors.shape
    if len(shape) == 1:
        s0 = shape[0]
        if s0 == 3:
            coors3D = coors.copy()
        elif s0 < 3:
            coors3D = np.zeros((3,))
            coors3D[:s0] = coors
        else:
            raise ValueError('Wrong coordinates shape %s' % coors.shape)
    elif len(shape) == 2:
        s0 = shape[0]
        s1 = shape[1]
        if s1 == 3:
            coors3D = coors.copy()
        elif s1 < 3:
            coors3D = np.zeros((s0, 3))
            coors3D[:, :s1] = coors
        else:
            raise ValueError('Wrong coordinates shape %s' % coors.shape)
    else:
        raise ValueError('Wrong coordinates shape %s' % coors.shape)
    return coors3D


def remove_dimensions(coors, tol=1e-7):
    coors = np.asarray(coors)
    assert len(coors.shape) == 2
    rows, cols = coors.shape
    assert cols <= 3
    cid = []
    for c in range(cols):
        if not (abs(coors[:, c] - coors[0, c]) < tol).all():
            cid.append(c)
    if cid == []:
        raise ValueError("Single point in coors")
    return coors[:, cid]


def add_dimensions(coors):
    coors = np.asarray(coors)
    rows, cols = coors.shape
    empty = np.zeros((rows, 3-cols))
    return np.append(coors, empty, axis=1)


def reduce_coors(coors, dmin=1e-7):
    from scipy.spatial.distance import pdist, squareform
    dist = squareform(pdist(coors))
    i, j = np.nonzero((dist < dmin))  # too close coors indices
    i = i[i > j]  # indices to remove
    mask = np.ones(len(coors), dtype=bool)
    mask[i] = False
    coors = coors[mask]
    return coors


def rank(coors, tol=1e-7):
    coors = coors3D(coors)
    center = coors.mean(axis=0)
    return np.linalg.matrix_rank(coors - center, tol=tol)


def coplanar(coors, tol=1e-7):
    if rank(coors, tol) == 2:
        return True
    return False


def colinear(coors, tol=1e-7):
    if rank(coors, tol) == 1:
        return True
    return False


#def contains_points(polygon, points, tol=1e-7):
#    ### polygon should be counterclockwise as matplotlib tests works better
#    polygon = coors3D(polygon)
#    points = coors3D(points)
#    if coplanar(np.vstack((polygon, points))):
#        from matplotlib.path import Path
#        for plane in [(0, 1), (0, 2), (1, 2)]:
#            poly = polygon[:, plane]
#            pts = points[:, plane]
#            if not colinear(poly, tol):
#                p = Path(poly).contains_points(pts)
#                return p
#    raise ValueError('Polygon and points are not coplanar')


def contains(polygon, points, tol=1e-7):
    # polygon should be counterclockwise as matplotlib tests works better
    polygon = coors3D(polygon)
    points = coors3D(points)
    if coplanar(np.vstack((polygon, points))):
        from shapely.geometry import Polygon, MultiPoint
        for plane in [(0, 1), (0, 2), (1, 2)]:
            poly = polygon[:, plane]
            pts = points[:, plane]
            if not colinear(poly, tol):
                p = Polygon(poly).contains(MultiPoint(pts))
                return p
    raise ValueError('Polygon and points are not coplanar')


def disjoint(polygon, points, tol=1e-7):
    # polygon should be counterclockwise as matplotlib tests works better
    polygon = coors3D(polygon)
    points = coors3D(points)
    if coplanar(np.vstack((polygon, points))):
        from shapely.geometry import Polygon, MultiPoint
        for plane in [(0, 1), (0, 2), (1, 2)]:
            poly = polygon[:, plane]
            pts = points[:, plane]
            if not colinear(poly, tol):
                p = Polygon(poly).disjoint(MultiPoint(pts))
                return p
    raise ValueError('Polygon and points are not coplanar')


def multi_linspace(starts, stops, **kwargs):
    assert len(starts) == len(stops)  # starts and stops should be equal in len
    space = []
    for st, sp in zip(starts, stops):
        space += [np.linspace(st, sp, **kwargs)]
    return np.asarray(space).T


def signed_angle(v1, v2):
    dot = np.dot(v1, v2)  # (v1 * v2).sum()
    slen = np.linalg.norm(v1) * np.linalg.norm(v2)
    if slen == 0:
        raise ValueError("Zero length vector.")
    angle = np.arccos(round(dot/slen, 15))
    cross = np.cross(v1, v2)
    sign = np.sign(cross.sum())  # cross.sum(1): dot product with [1,1,1]
    if not sign:
        sign = np.sign(dot)
    return angle*sign


def rotation_matrix(axis, angle):
    axis = coors3D(axis)
    ux, uy, uz = axis/np.linalg.norm(axis)
    c = np.cos(angle)  # angle in radians
    s = np.sin(angle)
    R = np.zeros((3, 3))
    R[0, 0] = c + (ux**2)*(1-c)
    R[0, 1] = ux*uy*(1-c) - uz*s
    R[0, 2] = ux*uz*(1-c) + uy*s
    R[1, 0] = uy*ux*(1-c) + uz*s
    R[1, 1] = c + (uy**2)*(1-c)
    R[1, 2] = uy*uz*(1-c) - ux*s
    R[2, 0] = uz*ux*(1-c) - uy*s
    R[2, 1] = uz*uy*(1-c) + ux*s
    R[2, 2] = c + (uz**2)*(1-c)
    return R


def uniquify(seq, idfun=None):
    # uses hash instesd of cmp so user defined __eq__ will not work
    # order preserving
    if idfun is None:
        def idfun(x):
            return x
    seen = {}
    result = []
    for item in seq:
        marker = idfun(item)
        # in old Python versions:
        # if seen.has_key(marker)
        # but in new ones:
        if marker in seen:
            continue
        seen[marker] = 1
        result.append(item)
    return result


#def uniquify(seq):
     # uses __eq__ but this is O(n^2), slow on large lists
     # order preserving
#    useq = []
#    for i in seq:
#        if i not in useq:
#            useq.append(i)
#    return useq


def tmpfname(ext=''):
    from tempfile import NamedTemporaryFile
    f = NamedTemporaryFile()
    fname = f.name + ext
    f.close()
    return fname


def saveobj(obj, fname, ext=''):
    from fempy.compat import pickle
    if ext:
        if not ext.startswith('.') and not fname.endswith('.'):
            ext = '.' + ext
        if not fname.endswith(ext):
            fname = fname + ext
    f = open(fname, 'wb')
    pickle.dump(obj, f)
    f.close()
    return


def loadobj(fname, ext=''):
    from fempy.compat import pickle
    if ext:
        if not ext.startswith('.') and not fname.endswith('.'):
            ext = '.' + ext
        if not fname.endswith(ext):
            fname = fname + ext
    f = open(fname, 'rb')
    obj = pickle.load(f)
    f.close()
    return obj


class CubicSpline(object):
    def eval(self, xn, diff=0):
        n = self.n
        x = self.x
        coeff = self.coeff
        ni = len(xn)
        ci = np.zeros(ni, dtype=int)  # np.zeros((ni, 4))
        xi = np.zeros(ni)
        for i in range(n-1):
            cdn = (x[i] <= xn) * (xn <= x[i+1])
            ci = np.where(cdn, i, ci)
            xi = np.where(cdn, 1-(x[i+1]-xn)/(x[i+1]-x[i]), xi)
        cn = coeff[ci]
        a, b, c, d = cn.T
        if diff == 0:
            return a*xi**3 + b*xi**2 + c*xi + d
        elif diff == 1:
            return 3*a*xi**2 + 2*b*xi + c
        elif diff == 2:
            return 6*a*xi + 2*b
        elif diff == 3:
            return 6*a
        else:
            raise ValueError("Wrong derivative order: %i" % (diff, ))


class NaturalSpline(CubicSpline):
    def __init__(self, y, closed=False):
        """
        Cubic spline defined for equidistant points in range (0,1).
        This is natural cubic spline, i.e. the first derivatives are
        calculated from continuous curvature condition. Curvature equal 0
        is assumed at boundaries if spline is not closed.

        See: http://mathworld.wolfram.com/CubicSpline.html
        """
        n = len(y)
        x = np.linspace(0, 1, n)
        y = np.array(y)
        dy = np.diff(y)

        # Derivatives
        Y = np.zeros(len(x))
        if not closed:
            Y[0] = y[1] - y[0]
            Y[1:-1] = y[2:] - y[:-2]
            Y[-1] = y[-1] - y[-2]
            Y = 3*Y
            # tridiagonal matrix
            A = np.diag(4*np.ones(n)) + \
                np.diag(np.ones(n-1), -1) + \
                np.diag(np.ones(n-1), 1)
            A[0, 0] = 2
            A[-1, -1] = 2
            D = np.linalg.solve(A, Y)
        else:
            Y[0] = y[1] - y[-2]
            Y[1:-1] = y[2:] - y[:-2]
            Y[-1] = y[1] - y[-2]
            Y = 3*Y
            # tridiagonal matrix
            A = np.diag(4*np.ones(n)) + \
                np.diag(np.ones(n-1), -1) + \
                np.diag(np.ones(n-1), 1)
            A[0, -2] = 1
            A[-1, 1] = 1
            D = np.linalg.solve(A, Y)

        # Coeficients
        a = -2*dy + D[:-1] + D[1:]
        b = 3*dy - 2*D[:-1] - D[1:]
        c = D[:-1]
        d = y[:-1]

        self.n = n
        self.x = x
        self.y = y
        self.coeff = np.vstack((a, b, c, d)).T


class GmshSpline(CubicSpline):
    def __init__(self, y, closed=False):
        """
        Cubic spline defined for equidistant points in range (0,1).
        This is *not* natural cubic spline, i.e. the first derivatives are
        not calculated from continuous curvature condition. Instead they
        are calculated directly from control points.
        """
        n = len(y)
        x = np.linspace(0, 1, n)
        y = np.array(y)
        dy = np.diff(y)

        # Derivatives
        D = np.zeros(n)
        D[1:-1] = (dy[:-1] + dy[1:])/2.
        if closed:
            D[0] = D[-1] = (dy[0] + dy[-1])/2.
        else:
            D[0] = dy[0]
            D[-1] = dy[-1]

        # Coeficients
        a = -2*dy + D[:-1] + D[1:]
        b = 3*dy - 2*D[:-1] - D[1:]
        c = D[:-1]
        d = y[:-1]

        self.n = n
        self.x = x
        self.y = y
        self.coeff = np.vstack((a, b, c, d)).T


# Tests
if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')



# -*- coding: utf-8 -*-
import pytest
from fempy.geometry import gmsh_io


# if __name__ == "__main__":
class Test_Dump_and_Preview(object):
    def test_2D_model_preview(self):
        from fempy.geometry import Rectangle
        from fempy.models.elastic import PlaneStrain

        # 2D
        d = Rectangle(elsize=0.025).gmsh.domain(order=1)
        m = PlaneStrain(d)
        m.ev(20000, 0.3)
        # gravity
        m.rho[:] = 2.
        m.gravity[:] = [0., -9.81]
        # pressures
        m.qi['right'] = [-10, 0.]
        m.p['left'] = -10.
        # dirichlet bcs
        m.u['top', 1] = -0.0005
        m.bcs['top', 1] = True
        m.bcs['bottom'] = True

        m.solve()
        # m.preview()
        data, nfields = gmsh_io.prepare_model_data(m, 'all')
        gmsh_io._preview(*data + ('-parse_and_exit',))  # no display

    def test_3D_model_preview(self):
        pytest.importorskip("pygmsh")
        from fempy.models.elastic import LinearElastic
        # 3D
        from fempy.geometry import Gmsh
        from pygmsh.built_in import Geometry
        pg = Geometry()
        pg.add_box(0, 0.3, 0, 0.1, 0, 0.1, 0.0125)
        g = Gmsh(pg, entities={2: {19: 'right',
                                   20: 'top',
                                   21: 'front',
                                   22: 'back',
                                   23: 'bottom',
                                   24: 'left'}})
        d = g.domain(order=1)
        m = LinearElastic(d)
        m.ev(20000, 0.3)
        # gravity
        m.rho[:] = 2.
        m.gravity[:] = [0., 0., -9.81]
        # pressures
        m.qi['right'] = [-10, 0., 0.]
        m.p['left'] = -10.
        # dirichlet bcs
        m.u['top', 1] = -0.0005
        m.bcs['top', 1] = True
        m.bcs['bottom'] = True

        m.solve()
        data, nfields = gmsh_io.prepare_model_data(m, 'all')
        gmsh_io._preview(*data + ('-parse_and_exit',))  # no display
        # m.preview_dump('test.msh', 'all', ascii=True)


if __name__ == '__main__':
    pytest.main([str(__file__), '-v'])  # Run tests from current file


# -*- coding: utf-8 -*-
"""
Created on Tue Jun 28 12:54:09 2016

@author: mwojc
"""
from __future__ import print_function
from fempy.compat import reload
import pytest
import inspect
import numpy as np
import fempy.solvers.methods as methods
import fempy.solvers.precon as precon
import fempy.solvers.nonlin as nonlin
from fempy.testing import testdata
from solvers import test_linear


def get_solvers():
    reload(methods)
    reload(precon)
    reload(nonlin)
    mc = [ls for ls in methods.__dict__.values()
          if inspect.isclass(ls) and issubclass(ls, methods.linear)]

    pc = [ls for ls in precon.__dict__.values()
          if inspect.isclass(ls) and issubclass(ls, precon.preconditioner)]

#    lsc = [ls for ls in linear.__dict__.values()
#           if inspect.isclass(ls) and issubclass(ls, linear.LinearSolver)]

    nsc = [ls for ls in nonlin.__dict__.values()
           if inspect.isclass(ls) and issubclass(ls, nonlin.NonLinearSolver)]
    return mc, pc, nsc


# TEST NONLIN
# howto INITIALIZE MODEL, OPTIONS, PRECONDITIOENR, ETc.


class Mscipy(object):
    info = "MODEL 1 - integro-differential equation from scipy docs"
    blacklist = ['SimpleIterationSolver']
    pblacklist = {}

    def __init__(self):
        n = 25
        self.n = n
        self.rhs = np.zeros((n, n))
        self.lhs = np.zeros((n, n))
        self.lhs0 = np.loadtxt(testdata + '/mscipy.sol').ravel()

    @property
    def lop(self):
        # parameters
        nx, ny = self.n, self.n
        hx, hy = 1./(nx-1), 1./(ny-1)

        P_left, P_right = 0, 0
        P_top, P_bottom = 1, 0

        P = self.lhs
        d2x = np.zeros_like(P)
        d2y = np.zeros_like(P)

        d2x[1:-1] = (P[2:]   - 2*P[1:-1] + P[:-2]) / hx/hx
        d2x[0]    = (P[1]    - 2*P[0]    + P_left)/hx/hx
        d2x[-1]   = (P_right - 2*P[-1]   + P[-2])/hx/hx
    
        d2y[:,1:-1] = (P[:,2:] - 2*P[:,1:-1] + P[:,:-2])/hy/hy
        d2y[:,0]    = (P[:,1]  - 2*P[:,0]    + P_bottom)/hy/hy
        d2y[:,-1]   = (P_top   - 2*P[:,-1]   + P[:,-2])/hy/hy

        return d2x + d2y - 10*np.cosh(P).mean()**2


class Test_NonLinear_Solvers(test_linear.Solver_Test):
    def test_nonlinear_solvers(self, capsys):
        N = 1  # NUMBER OF SOLVER PASSAGES FOR EVERY MODEL
#        test_linear.Mlm.pblacklist={'bicgstab': ['amg_sas', 'diag', 'lu',
#                                                 'inexact', 'exact', 'ilu',
#                                                 'preconditioner', 'umfpack',
#                                                 'none']}
        models = [test_linear.M, test_linear.Mbcs, test_linear.Mbcseq,
                  test_linear.Mlm, Mscipy]
        with capsys.disabled():
            print("\n")

            # First test NonLinearSolver against all available methods
            # and precons
            mc, pc, nsc = get_solvers()
            solvers = [nonlin.NonLinearSolver]
            methods = mc
            preconditioners = pc + [None]
            self.iterate_solvers(models, solvers, methods, preconditioners, N,
                                 f_tol=1e-8)

            # And then test the rest with default methods and preconditioners
            mc, pc, nsc = get_solvers()
            solvers = nsc
            methods = []
            preconditioners = []
            self.iterate_solvers(models, solvers, methods, preconditioners, N,
                                 f_tol=1e-8)


if __name__ == '__main__':
    # import pytest
    pytest.main([str(__file__), '-v'])  # Run tests from current file



# -*- coding: utf-8 -*-
"""
Created on Fri Apr  8 11:51:44 2016

@author: mwojc
"""
import scipy.sparse as sparse
import scipy.sparse.linalg
from fempy.solvers._solver import Solver


class linear(Solver):
    direct = False
    method = staticmethod(sparse.linalg.lgmres)

    def solve(self, A, b, x0=None, M=None, **options):
        options = self.getopts(options)
        x, info = self.method(A, b, x0=x0, M=M, **options)
        self.info = info
        return x, info
    __call__ = solve


# ITERATIVE SOLVERS
class iterative(linear):
    pass


class lgmres(iterative):
    pass


class gmres(iterative):
    method = staticmethod(sparse.linalg.gmres)

# Seems to not work without preconditioner
#class cg(iterative):
#    method = staticmethod(sparse.linalg.cg)


class cgs(iterative):
    method = staticmethod(sparse.linalg.cgs)


class bicgstab(iterative):
    method = staticmethod(sparse.linalg.bicgstab)


# This is solver only for symmetric systems - need to be tested
# class Minres(IterativeSolver):
#     method = staticmethod(sparse.linalg.minres)


# if pyamg is present
try:
    from pyamg import krylov
# Seems to not work if the system is not positively defined
#    class AmgCg(IterativeSolver):
#        method = staticmethod(krylov.cg)

# Seems to not work at all
#    class AmgSteepestDescent(IterativeSolver):
#        method = staticmethod(krylov.steepest_descent)

# This is solver only for symmetric systems - need to be tested
#    class AmgMinimalResidual(IterativeSolver):
#        method = staticmethod(krylov.minimal_residual)

# There are some exactness troubles in this solver it finishes before
# tol is reached
#    class amg_bicgstab(iterative):
#        method = staticmethod(krylov.bicgstab)

# Prints annoying warnings in testing
#    class amg_fgmres(iterative):
#        method = staticmethod(krylov.fgmres)

#    class amg_gmres(iterative):
#        method = staticmethod(krylov.gmres)
except ImportError:
    pass


## if pyamgcl is present
#try:
#    from pyamgcl import solver
#
#    class cg_amgcl(iterative):        
#        def method(self, A, b, x0=None, M=None, **options):
#            options = self.getopts(options)
#            options['type'] = 'cg'
#            S = solver(M, options)   # Would be inefficient to create solver
#            x = S(b)
#            return x, S.error
#except ImportError:
#    pass


# DIRECT SOLVER
class direct(linear):
    direct = True

    def method(self, A, b, x0=None, M=None, **options):
        if M is None:
            raise ValueError("No jacobian? Direct solver cannot handle this.")
        x = M*b
        return x, 0


if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')
#    import numpy as np
#    A = np.random.rand(5,5)
#    b = np.random.rand(5)
#    s = iterative()
#    print s.solve(A, b)



# -*- coding: utf-8 -*-
"""
Created on Wed May 18 12:07:58 2016

@author: mwojc
"""
from fempy.models.model import Model
from fempy.fields.nfields import nVector, nScalar
from fempy.fields.gfields import gField, gScalar, gVector
from fempy.fields.gfields import gTensor, gTensor3, gTensor4
from fempy.fields.operators import dot, grad, operator, onboundary, nabla
from fempy.fields.operators import qform, lform, integrate, ongauss
from fempy.fields.operators import BaseOperator
from fempy.fields.algebra import transpose, tensordot
import fempy.solvers.linear as linear
import numpy as np


def lame2cijkl(lam, mu, ndime=3):
    """
    Works for 3D and plain strain
    """
    lam = np.asarray(lam, dtype=np.float)
    mu = np.asarray(mu, dtype=np.float)
    assert lam.shape == mu.shape
    cijkl = np.zeros(lam.shape + (ndime,)*4)
    for i in range(ndime):
        for j in range(ndime):
            cijkl[..., i, i, j, j] += lam
            cijkl[..., j, i, j, i] += mu
            cijkl[..., j, i, i, j] += mu
    return cijkl


def ev2cijkl(E, v, ndime=3):
    """
    Works for 3D and plain strain
    """
    E = np.asarray(E, dtype=np.float)
    v = np.asarray(v, dtype=np.float)
    assert E.shape == v.shape
    lam = (E*v)/((1.+v)*(1.-2.*v))
    mu = E/(2*(1.+v))
    return lame2cijkl(lam, mu, ndime)


class LinearElastic(Model):
    default_solver = linear.DirectSolver

    def init(self, **kwargs):
        d = self.domain
        b = self.domain.boundary
        # on domain nodes
        self.u = nVector(d)             # displacement
        self.f = nVector(d)             # explicit right hand side forces
        self.bcs = nVector(d, False)    # Dirichlet boundary conditions
        # on domain integration points
        self.cijkl3D = gTensor4(d)      # material tensor
        self.aij3D = gTensor(d)         # thermal expansion [1/K]
        self.rho = gScalar(d)           # material density
        self.gravity = gVector(d)       # gravity vector
        self.gravity[..., -1] = -9.80655
        self.Tref = gScalar(d, 293.15)  # reference temperature
        self.T = gScalar(d, 293.15)     # current temperature
        # on boundary integration points
        self.p = gScalar(b)             # normal pressures on boundary
        self.t = gScalar(b)             # tangential pressures on boundary
        if d.ndime == 3:
            self.t = gField(b, [0, 0.])
        self.pij = gTensor(b)           # boundary stresses - element CS
        self.pi = gVector(b)            # boundary surface forces - elemet CS
        self.qi = gVector(b)            # boundary surface forces - global CS

        # displacement is our unknown
        self.lhs = self.u

    def lame(self, lam, mu, egroups=None):
        cijkl = lame2cijkl(lam, mu)
        if egroups is None:
            self.cijkl3D[:] = cijkl
        else:
            self.cijkl3D[egroups] = cijkl

    def ev(self, E, v, egroups=None):
        cijkl = ev2cijkl(E, v)
        if egroups is None:
            self.cijkl3D[:] = cijkl
        else:
            self.cijkl3D[egroups] = cijkl

    def alpha(self, alpha, egroups=None):
        aij = np.eye(3) * alpha
        if egroups is None:
            self.aij3D[:] = aij
        else:
            self.aij3D[egroups] = aij

    @property
    def cijkl(self):
        return self.cijkl3D

    @property
    def aij(self):
        return self.aij3D

    @property
    def rhs(self):
        u = self.u
        b = self.domain.boundary
        # Assemble body forces - weight
        weight = dot(self.rho, self.gravity)
        fw = lform(weight, operator(u))
        # Assemble boundary surface forces
        qi = dot(self.pij, b.normal)
        qi += dot(b.normal, self.p)
        qi += dot(b.tangent, self.t)
        qi += dot(transpose(b.uvw), self.pi)
        qi += self.qi
        fb = lform(qi, operator(onboundary(u)))
        # Assemble thermal load
        fdT = integrate(self.sigma_dT)  # should be lform(sigma_dT, bu)
        return self.f + fw + fb + fdT

    @property
    def jac(self):
        # Assemble stiffness matrix
        u = self.u
        nu = nabla(u)
        bu = 0.5*(nu + transpose(nu))     # small strain operator
        return qform(self.cijkl, bu, bu)  # the same trial and test

    @property
    def eps(self):
        du = grad(self.u)
        eps = 0.5*(du + transpose(du))
        return eps

    @property
    def rot(self):
        du = grad(self.u)
        rot = 0.5*(du - transpose(du))
        return rot

    @property
    def dT(self):
        return self.T - self.Tref

    @property
    def eps_dT(self):
        return dot(self.aij, self.dT)

    @property
    def sigma_dT(self):
        sigma_dT = tensordot(self.cijkl, self.eps_dT)
        return sigma_dT

    @property
    def sigma(self):
        sigma = tensordot(self.cijkl, self.eps - self.eps_dT)
        return sigma

    @property
    def lop(self):
        sigma_tot = tensordot(self.cijkl, self.eps)
        return integrate(sigma_tot)


class LinearElastic3D(LinearElastic):
    def init(self, *args, **kwargs):
        nd = self.domain.ndime
        ed = self.domain.etype.ndime
        assert nd == 3 and ed == 3, "3D domain is required"
        LinearElastic.init(self, **kwargs)


class LinearElastic2D(LinearElastic):
    def init(self, *args, **kwargs):
        nd = self.domain.ndime
        ed = self.domain.etype.ndime
        assert nd == 2 and ed == 2, "2D domain is required"
        LinearElastic.init(self, **kwargs)
        self.cijkl3D = gField(self.domain, np.zeros((3, 3, 3, 3)))
        self.aij3D = gField(self.domain, np.zeros((3, 3)))

    @property
    def cijkl(self):
        return gTensor4(self.domain, self.cijkl3D[..., :2, :2, :2, :2])

    @property
    def aij(self):
        return gTensor(self.domain, self.aij3D[..., :2, :2])

    @property
    def c2222(self):
        return gScalar(self.domain, self.cijkl3D[..., 2, 2, 2, 2])

    @property
    def c0022(self):
        return gScalar(self.domain, self.cijkl3D[..., 0, 0, 2, 2])

    @property
    def c1122(self):
        return gScalar(self.domain, self.cijkl3D[..., 1, 1, 2, 2])

    @property
    def aij22(self):
        return gScalar(self.domain, self.aij3D[..., 2, 2])


class PlaneStrain(LinearElastic2D):
    @property
    def sigma_dT(self):
        sigma_dT = super(PlaneStrain, self).sigma_dT
        sigma_dT[..., 0, 0] += self.c0022 * self.aij22 * self.dT
        sigma_dT[..., 1, 1] += self.c1122 * self.aij22 * self.dT
        return sigma_dT

    @property
    def sigmaz(self):
        eps = self.eps - self.eps_dT
        c0011 = self.cijkl[..., 0, 0, 1, 1]
        c0000 = self.cijkl[..., 0, 0, 0, 0]
        eps00 = eps[..., 0, 0]
        eps11 = eps[..., 1, 1]
        sig33 = c0011 * (eps00 + eps11) - c0000*self.eps_dT[..., 0, 0]
        return gScalar(self.domain, sig33)


class PlaneStress(LinearElastic2D):
    @property
    def cijkl(self):
        cijkl = gTensor4(self.domain, self.cijkl3D[..., :2, :2, :2, :2].copy())
        c0011 = cijkl[..., 0, 0, 1, 1]
        c0000 = cijkl[..., 0, 0, 0, 0]
        corr = (c0011**2)/c0000
        cijkl[..., 0, 0, 0, 0] -= corr
        cijkl[..., 0, 0, 1, 1] -= corr
        cijkl[..., 1, 1, 0, 0] -= corr
        cijkl[..., 1, 1, 1, 1] -= corr
        return cijkl

    @property
    def epsz(self):
        eps_dT = self.eps_dT
        eps = self.eps - eps_dT
        c0000 = self.cijkl[..., 0, 0, 0, 0]
        c0011 = self.cijkl[..., 0, 0, 1, 1]
        eps00 = eps[..., 0, 0]
        eps11 = eps[..., 1, 1]
        eps33 = -(c0011/c0000) * (eps00 + eps11) + eps_dT[..., 0, 0]  # isotropic
        return gScalar(self.domain, eps33)


class Axisymmetry(PlaneStrain):
    def init(self, **kwargs):
        PlaneStrain.init(self, **kwargs)
        axis = kwargs.pop('axis', 'auto')
        self.domain.axis = axis
        self.bcs['axis', 0] = True  # fix horizontal disp

    @property
    def radius(self):
        return self.domain.radius

    @property
    def epsz(self):
        epsz = ongauss(self.u)[..., 0] / self.radius
        return gScalar(self.domain, epsz)

    @property
    def sigmaz(self):
        eps = self.eps
        epsx = eps[..., 0, 0]
        epsy = eps[..., 1, 1]
        epsz = self.epsz
        sigmaz = self.cijkl[..., 0, 0, 1, 1] * (epsx + epsy) + \
            self.cijkl[..., 1, 1, 1, 1] * epsz
        return gScalar(self.domain, sigmaz)

    @property
    def sigma(self):
        sigma = tensordot(self.cijkl, self.eps)
        dsigma = self.cijkl[..., 0, 0, 1, 1] * self.epsz
        sigma[..., 0, 0] += dsigma
        sigma[..., 1, 1] += dsigma
        return sigma

    @property
    def lop(self):
        # Integrate in-plane stresses
        sigma = self.sigma
        integral = integrate(sigma)
        # Must add sigmaz contribution to first equation
        integralz = lform(self.sigmaz/self.radius)
        integral[..., 0] += integralz
        return integral

    @property
    def jac(self):
        # Plane strain part (non-linar solver works with this part only :))
        u = self.lhs
        nu = nabla(u)
        bu = 0.5*(nu + transpose(nu))     # small strain operator
        K1 = qform(self.cijkl, bu, bu)
        # Contribution of epsz on sigmax and sigmay and symmetrically
        # contribution of epsx and epsy on sigmaz
        r = self.radius
        ou = operator(u)[..., 0]  # radial displacement operator (u_0)
        ou = BaseOperator(u, ou)  # TODO: rethink general operator creation
        c0011 = self.cijkl[..., 0, 0, 1, 1]/r     # lambda/r
        dij = gTensor(self.domain)
        dij[..., 0, 0] = c0011
        dij[..., 1, 1] = c0011
        K2 = qform(dij, bu, ou)
        K3 = qform(dij, ou, bu)
        # Contribution of eps22 on sigma22
        c1111 = self.cijkl[..., 1, 1, 1, 1]/r**2  # (lambda + 2*mu)/r**2
        h = gScalar(self.domain, c1111)
        K4 = qform(h, ou, ou)
        # Sum and return
        # Note: resulting jacobian is not symmetric anymore...
        return K1 + K2 + K3 + K4


if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')



# -*- coding: utf-8 -*-
import os
import pytest
import tempfile
import numpy as np
from fempy.geometry import *
import fempy.testing as testing

testdata = testing.testdata  # dir for test data
tempdir = tempfile.gettempdir()


# WRITE BETTER TESTS PLEASE!
class TestGmsh:
    def test_save_read(self):
        #from fempy.geometry import Rectangle
        R = Rectangle()
        R.gmsh.save(tempdir + '/rectangle')
        GR = Gmsh((tempdir + '/rectangle.geo', tempdir + '/rectangle.msh'))
        # GR.preview()
        GR = Gmsh(tempdir + '/rectangle.geo')
        # GR.preview()
        GR = Gmsh(tempdir + '/rectangle.msh')
        # GR.preview()
        GR.domain()

    def test_geo_input(self):
        for name in os.listdir(testdata + '/gmsh'):
            n = testdata + '/gmsh/' + name
            if name.endswith('.geo'):
                gm = Gmsh(n)
                d = gm.domain()
                assert d.nelem > 0

    def test_msh_input(self):
        for name in os.listdir(testdata + '/gmsh'):
            n = testdata + '/gmsh/' + name
            if name.endswith('.geo'):
                gm = Gmsh(n)
                d1 = gm.domain()
                mn = tempdir + '/' + name[:-4] + '.msh'
                gm.save(mn)
                # re-read msh and test
                gm = Gmsh(mn)
                d2 = gm.domain()
                assert np.allclose(d1.conec, d2.conec)
            elif name.endswith('.msh'):
                gm = Gmsh(n)
                d = gm.domain()
                assert d.nelem > 0

    def test_brep_input(self):
        for name in os.listdir(testdata + '/gmsh'):
            n = testdata + '/gmsh/' + name
            if name.endswith('.brep'):
                gm = Gmsh(n)
                d = gm.domain()
                assert d.nelem > 0

    def test_iges_input(self):
        for name in os.listdir(testdata + '/gmsh'):
            n = testdata + '/gmsh/' + name
            if name.endswith('.iges'):
                gm = Gmsh(n)
                d = gm.domain()
                assert d.nelem > 0

    def test_pygmsh_input(self):
        pytest.importorskip("pygmsh")
        from fempy.models.elastic import LinearElastic
        # 3D
        from fempy.geometry import Gmsh
        from pygmsh.built_in import Geometry
        pg = Geometry()
        pg.add_box(0, 0.3, 0, 0.1, 0, 0.1, 0.025)
        g = Gmsh(pg, entities={2: {19: 'right',
                                   20: 'top',
                                   21: 'front',
                                   22: 'back',
                                   23: 'bottom',
                                   24: 'left'}})
        d = g.domain(order=1)
        assert d.nelem > 0

    def test_geometry_input(self):
        R = Rectangle()
        gm = Gmsh(R)
        d = gm.domain()
        assert d.nelem > 0

    def test_multiple_physical_names(self):
        d = testing.common.SquareWithCircle()
        assert np.allclose(d.vol, 1.)
        assert np.allclose(d.boundary.vol, 4.)
        assert d['r'].nelem + d['c'].nelem == d.nelem
        assert d['domain'].nelem == d.nelem


if __name__ == '__main__':
    pytest.main([str(__file__), '-v'])  # Run tests from current file



# -*- coding: utf-8 -*-
"""
Created on Tue Jun 28 11:17:33 2016

@author: mwojc
"""
from fempy.domain.domain import Domain
import fempy.models.elastic as elastic
from fempy.solvers.tools import model_test
from fempy.testing import testdata
import common
import numpy as np


class Test_Elastic(object):
    def test_mesh1_strain(self):
        coors = np.loadtxt(testdata + '/elastic/mesh1.coors')
        conec = np.loadtxt(testdata + '/elastic/mesh1.conec',
                           dtype=int)
        sol = np.loadtxt(testdata + '/elastic/mesh1.pstrain.sol')

        d = Domain(coors, conec)
        d.eidx.nidx.g._vars['tol'] = 1e-7
        m = elastic.PlaneStrain(d)
        # material
        m.ev(30000., 0.2)
        m.gravity[:] = [0., -9.80665]
        m.rho[:] = 20./9.80665
        # boundary conditions
        m.bcs['x<tol', 0] = True       # left
        m.bcs['x>2-tol', 0] = True     # right
        m.bcs['y<tol'] = True          # bottom
        # load
        m.p['(y>1-tol)*(x>1-tol)'] = -100.

        m.solve()
        # m.preview()

        assert np.allclose(m.lhs, sol)
        assert np.allclose(m.jac_times_lhs, m.lop)
        model_test(m, x0=sol)

    def test_mesh1_axisym(self):
        coors = np.loadtxt(testdata + '/elastic/mesh1.coors')
        conec = np.loadtxt(testdata + '/elastic/mesh1.conec',
                           dtype=int)
        sol = np.loadtxt(testdata + '/elastic/mesh1.axisym.sol')

        d = Domain(coors, conec)
        d.eidx.nidx.g._vars['tol'] = 1e-7
        m = elastic.Axisymmetry(d)
        # material
        m.ev(30000., 0.2)
        m.gravity[:] = [0., -9.80665]
        m.rho[:] = 20./9.80665
        # boundary conditions
        m.bcs['x<tol', 0] = True       # left
        m.bcs['x>2-tol', 0] = True     # right
        m.bcs['y<tol'] = True          # bottom
        # load
        m.p['(y>1-tol)*(x>1-tol)'] = -100.

        m.solve()
        # m.preview()

        assert np.allclose(m.lhs, sol)
        assert np.allclose(m.jac_times_lhs, m.lop)
        model_test(m, x0=sol)

    def test_mesh2_pstrain(self):
        coors = np.loadtxt(testdata + '/elastic/mesh2.coors')
        conec = np.loadtxt(testdata + '/elastic/mesh2.conec',
                           dtype=int)
        sol = np.loadtxt(testdata + '/elastic/mesh2.pstrain.sol')

        d = Domain(coors, conec)
        d.eidx.nidx.g._vars['tol'] = 1e-7
        d.eidx.nidx.g._vars['r'] = 2./2.
        m = elastic.PlaneStrain(d)
        m.ev(30000, 0.2)
        # boundary conditions
        m.bcs['(np.abs(x)<tol)*(np.abs(y)<r+tol)', 0] = True
        m.bcs['(np.abs(y)<tol)*(np.abs(x)<r+tol)', 1] = True
        # load
        m.p['(x<x.min()+tol)'] = 100.  # bottom
        m.p['(x>x.max()-tol)'] = 100.  # top

        m.solve()
        # m.preview()

        assert np.allclose(m.lhs, sol)
        assert np.allclose(m.jac_times_lhs, m.lop)
        model_test(m, x0=sol)

    def test_simple_pstrain(self):
        sol = np.loadtxt(testdata + '/elastic/simple.pstrain.sol')
        d = common.Simple2D().domain2Dmv([0.1, 0.1])
        m = elastic.PlaneStrain(d)
        m.ev(1000., 0.25)
        m.bcs['l', 0] = True
        m.bcs['c1'] = True
        m.p['r'] = 5.
        m.solve()
        # m.preview()
        assert np.allclose(m.lhs, sol, atol=1e-7)
        assert np.allclose(m.jac_times_lhs, m.lop)
        model_test(m, x0=sol)

    def test_simple_pstress(self):
        sol = np.loadtxt(testdata + '/elastic/simple.pstress.sol')
        d = common.Simple2D().domain2Dmv([0.1, 0.1])
        m = elastic.PlaneStress(d)
        m.domain.thickness[:] = 0.1
        m.domain.boundary.thickness[:] = 0.1
        m.ev(1000., 0.25)
        m.bcs['l', 0] = True
        m.bcs['c1'] = True
        m.p['r'] = 5.
        m.solve()
        # m.preview()
        assert np.allclose(m.lhs, sol)
        assert np.allclose(m.jac_times_lhs, m.lop)
        model_test(m, x0=sol)

    def test_simple_axisym(self):
        sol = np.loadtxt(testdata + '/elastic/simple.axisym.sol')
        d = common.Simple2D().domain2Dmv([0.1, 0.1])
        m = elastic.Axisymmetry(d)
        m.ev(1000., 0.25)
        m.bcs['l', 0] = True
        m.bcs['c1'] = True
        m.p['r'] = 5.
        m.solve()
        # m.preview()
        assert np.allclose(m.lhs, sol)
        assert np.allclose(m.jac_times_lhs, m.lop)
        model_test(m, x0=sol)

    def test_simple_3D(self):
        sol = np.loadtxt(testdata + '/elastic/simple.3D.sol')
        d = common.Simple3D().domain3D
        m = elastic.LinearElastic3D(d)
        m.ev(1000., 0.25)
        m.bcs['left', 1] = True
        m.bcs['bottom*left'] = True
        m.p['right'] = 5.
        m.solve()
#        m.preview()
        assert np.allclose(m.lhs, sol, atol=1e-7)
        assert np.allclose(m.jac_times_lhs, m.lop)
        model_test(m, x0=sol)

    def test_thermal_plane_strain(self):
        class Thermal3D(elastic.LinearElastic3D):
            def setup_domain(self):
                from fempy.geometry.shapes import Box
                self.geometry = Box(10, 1, 10, elsize=0.5)
                self.domain = self.geometry.gmsh.domain()

            def setup_model(self):
                self.bcs['left'] = [False, True, False]
                self.bcs['right'] = [False, True, False]
                self.bcs['back*left*bottom'] = [True, True, True]
                self.bcs['front*left*bottom'] = [False, True, True]
                self.ev(20000., 0.3)
                self.alpha(1e-4)
                self.T[:] += 3.

        class Thermal2DPlaneStrain(elastic.PlaneStrain):
            def setup_domain(self):
                from fempy.geometry.shapes import Rectangle
                self.geometry = Rectangle(10, 10, lb=(0, 0), elsize=0.5)
                self.domain = self.geometry.gmsh.domain()

            def setup_model(self):
                self.bcs['bottom_left'] = [True, True]
                self.bcs['bottom_right'] = [False, True]
                self.ev(20000., 0.3)
                self.alpha(1e-4)
                self.T[:] += 3.

        m3D = Thermal3D()
        m3D.solve()
#        from fempy.geometry.gmsh_io import preview_field
#        preview_field(m3D.sigma[..., 1, 1], name='sigmaz', d=m3D.domain)
        assert np.allclose(m3D.sigma[..., 1, 1], -6.)
        assert np.allclose(m3D.res.sum(), 0.)
        assert np.allclose(m3D.res['~(left+right)'], 0.)

        m2D = Thermal2DPlaneStrain()
        m2D.solve()
        # from fempy.geometry.gmsh_io import preview_field
        # preview_field(m2D.sigmaz, name='sigmaz', d=m2D.domain)
        assert np.allclose(m2D.sigmaz, -6.)
        assert np.allclose(m2D.res.sum(), 0.)
        assert np.allclose(m2D.res['~(bottom_left+bottom_right)'], 0.)

    def test_thermal_plane_strain_dist(self):
        class Thermal3D(elastic.LinearElastic3D):
            def setup_domain(self):
                from fempy.geometry.shapes import Box
                self.geometry = Box(10, 1, 10, elsize=0.5)
                self.domain = self.geometry.gmsh.domain(1)

            def setup_model(self):
                self.bcs['left'] = [False, True, False]
                self.bcs['right'] = [False, True, False]
                self.bcs['back*left*bottom'] = [True, True, True]
                self.bcs['front*left*bottom'] = [False, True, True]
                self.ev(20000., 0.3)
                self.alpha(1e-4)
                self.T[:] += 1*self.domain.gcoors[..., 0]

        class Thermal2DPlaneStrain(elastic.PlaneStrain):
            def setup_domain(self):
                from fempy.geometry.shapes import Rectangle
                self.geometry = Rectangle(10, 10, lb=(0, 0), elsize=0.5)
                self.domain = self.geometry.gmsh.domain(1)

            def setup_model(self):
                self.bcs['bottom_left'] = [True, True]
                self.bcs['bottom_right'] = [False, True]
                self.ev(20000., 0.3)
                self.alpha(1e-4)
                self.T[:] += 1*self.domain.gcoors[..., 0]

        m3D = Thermal3D()
        m3D.solve()
#        from fempy.geometry.gmsh_io import preview_field
#        preview_field(m3D.sigma[..., 1, 1], name='sigmaz', d=m3D.domain)
        assert np.allclose(m3D.res.sum(), 0.)
        assert np.allclose(m3D.res['~(left+right)'], 0.)

        m2D = Thermal2DPlaneStrain()
        m2D.solve()
#        from fempy.geometry.gmsh_io import preview_field
#        preview_field(m2D.sigmaz, name='sigmaz', d=m2D.domain)
        assert np.allclose(m2D.res.sum(), 0.)
        assert np.allclose(m2D.res['~(bottom_left+bottom_right)'], 0.)

        # Solutions should be close
        assert np.allclose(m2D.u.max(), m3D.u.max(), 4)

    def test_thermal_plane_stress(self):
        class Thermal3D(elastic.LinearElastic3D):
            def setup_domain(self):
                from fempy.geometry.shapes import Box
                self.geometry = Box(10, 1, 10, elsize=0.5)
                self.domain = self.geometry.gmsh.domain()

            def setup_model(self):
                # self.bcs['left'] = [False, True, False]
                # self.bcs['right'] = [False, True, False]
                self.bcs['back*left*bottom'] = [True, True, True]
                self.bcs['front*left*bottom'] = [False, False, True]
                self.bcs['back*right*bottom'] = [True, False, False]
                self.bcs['back*left*top'] = [False, True, False]
                # self.bcs['abc'] = [True, True, True]
                self.ev(20000., 0.3)
                self.alpha(1e-4)
                self.T[:] += 3.

        class Thermal2DPlaneStress(elastic.PlaneStress):
            def setup_domain(self):
                from fempy.geometry.shapes import Rectangle
                self.geometry = Rectangle(10, 10, lb=(0, 0), elsize=0.5)
                self.domain = self.geometry.gmsh.domain()

            def setup_model(self):
                self.bcs['bottom_left'] = [True, True]
                self.bcs['bottom_right'] = [False, True]
                self.ev(20000., 0.3)
                self.alpha(1e-4)
                self.T[:] += 3.

        m3D = Thermal3D()
        m3D.solve()
#        from fempy.geometry.gmsh_io import preview_field
#        preview_field(m3D.eps[..., 1, 1], name='epsz', d=m3D.domain)
        assert np.allclose(m3D.eps[..., 1, 1], 0.0003)

        m2D = Thermal2DPlaneStress()
        m2D.solve()
#        from fempy.geometry.gmsh_io import preview_field
#        preview_field(m2D.epsz, name='epsz', d=m2D.domain)
        assert np.allclose(m2D.epsz, 0.0003)


if __name__ == '__main__':
    import pytest
    pytest.main([str(__file__), '-v'])  # Run tests from current file



# -*- coding: utf-8 -*-
#
'''
I/O for Gmsh's msh format, cf.
<http://geuz.org/gmsh/doc/texinfo/gmsh.html#File-formats>.

.. moduleauthor:: Nico Schlömer <nico.schloemer@gmail.com>


This module is copied form meshio project and adapted for fempy needs
Added: dumpig $NodeData, examples

'''
import logging
import struct
import collections
import numpy

num_nodes_per_cell = {
    'vertex': 1,
    'line': 2,
    'triangle': 3,
    'quad': 4,
    'tetra': 4,
    'hexahedron': 8,
    'wedge': 6,
    'pyramid': 5,
    #
    'line3': 3,
    'triangle6': 6,
    'quad9': 9,
    'tetra10': 10,
    'hexahedron27': 27,
    'prism18': 18,
    'pyramid14': 14,
    'line4': 4,
    'quad16': 16,
    }

# Translate meshio types to gmsh codes
# http://geuz.org/gmsh/doc/texinfo/gmsh.html#MSH-ASCII-file-format
_gmsh_to_meshio_type = {
        15: 'vertex',
        1: 'line',
        2: 'triangle',
        3: 'quad',
        4: 'tetra',
        5: 'hexahedron',
        6: 'wedge',
        7: 'pyramid',
        8: 'line3',
        9: 'triangle6',
        10: 'quad9',
        11: 'tetra10',
        12: 'hexahedron27',
        13: 'prism18',
        14: 'pyramid14',
        26: 'line4',
        36: 'quad16',
        }
_meshio_to_gmsh_type = {v: k for k, v in _gmsh_to_meshio_type.items()}

_fempy_to_meshio_type = {
        'point': 'vertex',
        'line2': 'line',
        'line3': 'line3',
        'tria3': 'triangle',
        'tria6': 'triangle6',
        'tetr4': 'tetra',
        'tetr10': 'tetra10'
        }
_meshio_to_fempy_type = {v: k for k, v in _fempy_to_meshio_type.items()}


def read(filename):
    '''Reads a Gmsh msh file.
    '''
    with open(filename, 'rb') as f:
        out = read_buffer(f)
    return out


def read_buffer(f):
    # The format is specified at
    # <http://geuz.org/gmsh/doc/texinfo/gmsh.html#MSH-ASCII-file-format>.

    # Initialize the data optional data fields
    points = []
    cells = {}
    field_data = {}
    cell_data = {}
    point_data = {}

    has_additional_tag_data = False
    is_ascii = None
    int_size = 4
    data_size = None
    while True:
        line = f.readline().decode('utf-8')
        if not line:
            # EOF
            break
        assert line[0] == '$'
        environ = line[1:].strip()
        if environ == 'MeshFormat':
            line = f.readline().decode('utf-8')
            # Split the line
            # 2.2 0 8
            # into its components.
            str_list = list(filter(None, line.split()))
            assert str_list[0][0] == '2', 'Need mesh format 2'
            assert str_list[1] in ['0', '1']
            is_ascii = str_list[1] == '0'
            data_size = int(str_list[2])
            if not is_ascii:
                # The next line is the integer 1 in bytes. Useful to check
                # endianness. Just assert that we get 1 here.
                one = f.read(int_size)
                assert struct.unpack('i', one)[0] == 1
                line = f.readline().decode('utf-8')
                assert line == '\n'
            line = f.readline().decode('utf-8')
            assert line.strip() == '$EndMeshFormat'
        # TODO PhysicalNames code does not work properly
        elif environ == 'PhysicalNames':
            line = f.readline().decode('utf-8')
            num_phys_names = int(line)
            for _ in range(num_phys_names):
                line = f.readline().decode('utf-8')
#                key = line.split(' ')[2].replace('"', '').replace('\n', '')
#                phys_group = int(line.split(' ')[1])
#                field_data[key] = phys_group
            line = f.readline().decode('utf-8')
            assert line.strip() == '$EndPhysicalNames'
        # ####
        elif environ == 'Nodes':
            # The first line is the number of nodes
            line = f.readline().decode('utf-8')
            num_nodes = int(line)
            if is_ascii:
                points = numpy.fromfile(
                    f, count=num_nodes*4, sep=' '
                    ).reshape((num_nodes, 4))
                # The first number is the index
                points = points[:, 1:]
            else:
                # binary
                num_bytes = num_nodes * (int_size + 3 * data_size)
                assert numpy.int32(0).nbytes == int_size
                assert numpy.float64(0.0).nbytes == data_size
                dtype = [('index', numpy.int32), ('x', numpy.float64, (3,))]
                data = numpy.fromstring(f.read(num_bytes), dtype=dtype)
                assert (data['index'] == range(1, num_nodes+1)).all()
                # vtk numpy support requires contiguous data
                points = numpy.ascontiguousarray(data['x'])
                line = f.readline().decode('utf-8')
                assert line == '\n'

            line = f.readline().decode('utf-8')
            assert line.strip() == '$EndNodes'
        elif environ == 'Elements':
#            assert environ == 'Elements', \
#                'Unknown environment \'{}\'.'.format(environ)
            # The first line is the number of elements
            line = f.readline().decode('utf-8')
            total_num_cells = int(line)
            if is_ascii:
                for _ in range(total_num_cells):
                    line = f.readline().decode('utf-8')
                    data = [int(k) for k in filter(None, line.split())]
                    t = _gmsh_to_meshio_type[data[1]]
                    num_nodes_per_elem = num_nodes_per_cell[t]

                    if t not in cells:
                        cells[t] = []
                    cells[t].append(data[-num_nodes_per_elem:])

                    # data[2] gives the number of tags. The gmsh manual
                    # <http://gmsh.info/doc/texinfo/gmsh.html#MSH-ASCII-file-format>
                    # says:
                    # >>>
                    # By default, the first tag is the number of the physical
                    # entity to which the element belongs; the second is the
                    # number of the elementary geometrical entity to which the
                    # element belongs; the third is the number of mesh
                    # partitions to which the element belongs, followed by the
                    # partition ids (negative partition ids indicate ghost
                    # cells). A zero tag is equivalent to no tag. Gmsh and most
                    # codes using the MSH 2 format require at least the first
                    # two tags (physical and elementary tags).
                    # <<<
                    num_tags = data[2]
                    if t not in cell_data:
                        cell_data[t] = []
                    cell_data[t].append(data[3:3+num_tags])

                # convert to numpy arrays
                for key in cells:
                    cells[key] = numpy.array(cells[key], dtype=int)
                for key in cell_data:
                    cell_data[key] = numpy.array(cell_data[key], dtype=int)
            else:
                # binary
                num_elems = 0
                while num_elems < total_num_cells:
                    # read element header
                    elem_type = struct.unpack('i', f.read(int_size))[0]
                    t = _gmsh_to_meshio_type[elem_type]
                    num_nodes_per_elem = num_nodes_per_cell[t]
                    num_elems0 = struct.unpack('i', f.read(int_size))[0]
                    num_tags = struct.unpack('i', f.read(int_size))[0]
                    # assert num_tags >= 2

                    # read element data
                    num_bytes = 4 * (
                        num_elems0 * (1 + num_tags + num_nodes_per_elem)
                        )
                    shape = \
                        (num_elems0, 1 + num_tags + num_nodes_per_elem)
                    b = f.read(num_bytes)
                    data = numpy.fromstring(
                        b, dtype=numpy.int32
                        ).reshape(shape)

                    if t not in cells:
                        cells[t] = []
                    cells[t].append(data[:, -num_nodes_per_elem:])

                    if t not in cell_data:
                        cell_data[t] = []
                    cell_data[t].append(data[:, 1:num_tags+1])

                    num_elems += num_elems0

                # collect cells
                for key in cells:
                    cells[key] = numpy.vstack(cells[key])

                # collect cell data
                for key in cell_data:
                    cell_data[key] = numpy.vstack(cell_data[key])

                line = f.readline().decode('utf-8')
                assert line == '\n'

            line = f.readline().decode('utf-8')
            assert line.strip() == '$EndElements'

            # Subtract one to account for the fact that python indices are
            # 0-based.
            for key in cells:
                cells[key] -= 1

            # restrict to the standard two data items
            output_cell_data = {}
            for key in cell_data:
                if cell_data[key].shape[1] > 2:
                    has_additional_tag_data = True
                output_cell_data[key] = {}
                if cell_data[key].shape[1] > 0:
                    output_cell_data[key]['physical'] = cell_data[key][:, 0]
                if cell_data[key].shape[1] > 1:
                    output_cell_data[key]['geometrical'] = cell_data[key][:, 1]
            cell_data = output_cell_data
        else:
            logging.warning(
                'Environment \'{}\' is not supported.'.format(environ)
                )
            while not line.startswith('$End'):  # might be slow but is safe
                line = f.readline().decode('utf-8')
#            break  # immediate bu not safe


    if has_additional_tag_data:
        logging.warning(
            'The file contains tag data that couldn\'t be processed.'
            )

    return points, cells, point_data, cell_data, field_data


def write(
        filename,
        points,
        cells,
        is_ascii=False,
        point_data=None,
        cell_data=None,
        field_data=None
        ):
    '''Writes msh files, cf.
    http://geuz.org/gmsh/doc/texinfo/gmsh.html#MSH-ASCII-file-format
    '''
    point_data = {} if point_data is None else point_data
    cell_data = {} if cell_data is None else cell_data
    field_data = {} if field_data is None else field_data

    if not is_ascii:
        for key in cells:
            if cells[key].dtype != numpy.int32:
#                logging.warning(
#                    'Binary Gmsh needs 32-bit integers. Converting.'
#                    )
                cells[key] = numpy.array(cells[key], dtype=numpy.int32)
    mode = 'wb'    
    if is_ascii:
        mode = 'w'

    with open(filename, mode) as fh:
        mode_idx = 0 if is_ascii else 1
        size_of_double = 8
        fh.write((
            '$MeshFormat\n2.2 {} {}\n'.format(mode_idx, size_of_double)
            ).encode('utf-8'))
        if not is_ascii:  # binary
            fh.write(struct.pack('i', 1))
            fh.write('\n'.encode('utf-8'))
        fh.write('$EndMeshFormat\n'.encode('utf-8'))

        # Write nodes
        fh.write('$Nodes\n'.encode('utf-8'))
        fh.write('{}\n'.format(len(points)).encode('utf-8'))
        if is_ascii:
            for k, x in enumerate(points):
                fh.write(
                    '{} {!r} {!r} {!r}\n'.format(k+1, x[0], x[1], x[2])
                    .encode('utf-8')
                    )
        else:
            dtype = [('index', numpy.int32), ('x', numpy.float64, (3,))]
            tmp = numpy.empty(len(points), dtype=dtype)
            tmp['index'] = 1 + numpy.arange(len(points))
            tmp['x'] = points
            fh.write(tmp.tostring())
            fh.write('\n'.encode('utf-8'))
        fh.write('$EndNodes\n'.encode('utf-8'))

        fh.write('$Elements\n'.encode('utf-8'))
        # count all cells
        total_num_cells = sum([data.shape[0] for _, data in cells.items()])
        fh.write('{}\n'.format(total_num_cells).encode('utf-8'))

        consecutive_index = 0
        cell_numbers = {}
        for cell_type, node_idcs in cells.items():
            cell_numbers[cell_type] = \
                numpy.arange(consecutive_index,
                             consecutive_index + len(node_idcs))
            # handle cell data
            if cell_type in cell_data and cell_data[cell_type]:
                datas = []
                for key in sorted(cell_data[cell_type].keys()):
                    if key == 'physical':
                        ph = cell_data[cell_type]['physical']
                        assert len(ph) == len(node_idcs)
                    else:
                        ph = numpy.ones(len(node_idcs), dtype=numpy.int32)
                    if key == 'geometrical':
                        geo = cell_data[cell_type]['geometrical']
                        assert len(geo) == len(node_idcs)
                    else:
                        geo = numpy.ones(len(node_idcs), dtype=numpy.int32)
                    if key not in ('physical', 'geometrical'):
                        dat = cell_data[cell_type][key]
                        if len(dat) == len(node_idcs):  # not sufficient?
                            datas += [dat]
                    # TODO assert that the data type is int
                fcd = numpy.column_stack([ph, geo] + datas)
            else:
                # no cell data
                fcd = numpy.empty([len(node_idcs), 0], dtype=numpy.int32)

            if is_ascii:
                form = '{} ' + str(_meshio_to_gmsh_type[cell_type]) \
                    + ' ' + str(fcd.shape[1]) \
                    + ' {} {}\n'
                for k, c in enumerate(node_idcs):
                    fh.write(
                        form.format(
                            consecutive_index + k + 1,
                            ' '.join([str(val) for val in fcd[k]]),
                            ' '.join([str(cc + 1) for cc in c])
                            ).encode('utf-8')
                        )
            else:
                # header
                fh.write(struct.pack('i', _meshio_to_gmsh_type[cell_type]))
                fh.write(struct.pack('i', node_idcs.shape[0]))
                fh.write(struct.pack('i', fcd.shape[1]))
                # actual data
                a = numpy.arange(
                    len(node_idcs), dtype=numpy.int32
                    )[:, numpy.newaxis]
                a += 1 + consecutive_index
                array = numpy.hstack([a, fcd, node_idcs + 1])
                fh.write(array.tostring())

            consecutive_index += len(node_idcs)
        if not is_ascii:
            fh.write('\n'.encode('utf-8'))
        fh.write('$EndElements\n'.encode('utf-8'))

        # Write point data
        num_nodes = len(points)
        for data_name, data in point_data.items():
            for i, di in enumerate(data):
                time, arr = di
                n = arr.shape[0]
                arr = arr.reshape(n, -1)
                f = arr.shape[1]
                assert n == num_nodes
                assert f in (1, 3, 9)
                fh.write('$NodeData\n'.encode('utf-8'))
                fh.write('{}\n'.format('1').encode('utf-8'))
                fh.write(('"'+data_name+'"\n').encode('utf-8'))
                fh.write('{}\n'.format('1').encode('utf-8'))
                fh.write('{}\n'.format(str(float(time))).encode('utf-8'))
                fh.write('{}\n'.format('3').encode('utf-8'))
                fh.write('{}\n'.format(str(i)).encode('utf-8'))
                fh.write('{}\n'.format(str(f)).encode('utf-8'))
                fh.write('{}\n'.format(str(n)).encode('utf-8'))
                if not is_ascii:
                    dtype = [('index', numpy.int32),
                             ('field', numpy.float64, (f,))]
                    tmp = numpy.empty(n, dtype=dtype)
                    tmp['index'] = 1 + numpy.arange(n)
                    tmp['field'] = arr
                    fh.write(tmp.tostring())
                    fh.write('\n'.encode('utf-8'))
                else:
                    form = '{} '+'{}\n'*f
                    for nidx in range(n):
                        nf = [nidx+1] + arr[nidx].tolist()
                        fh.write(form.format(*nf).encode('utf-8'))
                fh.write('$EndNodeData\n'.encode('utf-8'))

        # Write element data
        for cell_type_i, cell_data_i in cell_data.items():
            cell_numbers_i = cell_numbers[cell_type_i]
            num_cells = len(cell_numbers_i)
            for data_name, data in cell_data_i.items():
                for i, di in enumerate(data):
                    time, arr = di
                    n = arr.shape[0]
                    arr = arr.reshape(n, -1)
                    f = arr.shape[1]
                    assert n == num_cells
                    assert f in (1, 3, 9)
                    fh.write('$ElementData\n'.encode('utf-8'))
                    fh.write('{}\n'.format('1').encode('utf-8'))
                    fh.write(('"'+data_name+'"\n').encode('utf-8'))
                    fh.write('{}\n'.format('1').encode('utf-8'))
                    fh.write('{}\n'.format(str(float(time))).encode('utf-8'))
                    fh.write('{}\n'.format('3').encode('utf-8'))
                    fh.write('{}\n'.format(str(i)).encode('utf-8'))
                    fh.write('{}\n'.format(str(f)).encode('utf-8'))
                    fh.write('{}\n'.format(str(n)).encode('utf-8'))
                    if not is_ascii:
                        dtype = [('index', numpy.int32),
                                 ('field', numpy.float64, (f,))]
                        tmp = numpy.empty(n, dtype=dtype)
                        tmp['index'] = 1 + cell_numbers_i
                        tmp['field'] = arr
                        fh.write(tmp.tostring())
                        fh.write('\n'.encode('utf-8'))
                    else:
                        form = '{} '+'{}\n'*f
                        for ni, nidx in enumerate(cell_numbers_i):
                            nf = [nidx+1] + arr[ni].tolist()
                            fh.write(form.format(*nf).encode('utf-8'))
                    fh.write('$EndElementData\n'.encode('utf-8'))
    return


# Fempy previewing and dumping functions
from fempy.geometry import tools
import subprocess, os, inspect


def prepare_domain_data(d, add_normals=True):
    coors = tools.add_dimensions(d.coors)
    conec = d.conec
    etype = _fempy_to_meshio_type[d.etype.name]

    # Meshio data
    points = coors
    cells = {etype: conec}  # , betype: bconec}
    point_data = {}
    cell_data = {}
    data = (points, cells, point_data, cell_data)
    if hasattr(d, 'normal') and add_normals:
        data = prepare_field_data(d.normal, 'domain_normals',
                                  d, data)
    # Handle boundary  (use recurency here?)
    if d.boundary is not None:
        if d.boundary.nelem > 0:
            # PLEASE DO SOMETHING WITH THIS MESS
            lst = d.boundary.eidx.nidx.lst
            pos = d.eidx.nidx.pos[lst]
            bconec = pos[d.boundary.conec]
            # PLEASE
            betype = _fempy_to_meshio_type[d.boundary.etype.name]
            cells[betype] = bconec
            if hasattr(d.boundary, 'normal') and add_normals:
                data = prepare_field_data(d.boundary.normal,
                                          'boundary_normals',
                                          d.boundary, data)
    return data


def prepare_field_data(f, name, d=None, data=None, time=0.):
    assert hasattr(f, 'domain') or d is not None
    d = f.domain if d is None else d  # preferes d

    if not hasattr(f, 'domain'):
        if len(f) == d.nnode:  # nField is assumed
            from fempy.fields.nfields import nField
            f = nField(d, f)
        elif len(f) == d.nelem:  # gField is assumed
            from fempy.fields.gfields import gField
            f = gField(d, f)
        else:
            raise ValueError("Given array cannot be converted to proper field")

    if data is None:
        data = prepare_domain_data(d, add_normals=False)
    else:
        pass  # consistency should be checked?

    # Shape tests
    if f.pshape == (d.nnode,):              # nField
        if f.sshape == ():                  # nScalar
            pass
        elif f.sshape == (d.ndime,):        # nVector
            f = tools.add_dimensions(f)
        elif f.sshape == (d.ndime,)*2:      # nTensor
            if d.ndime < 3:
                fz = numpy.zeros(f.pshape + (3, 3))
                fz[..., :d.ndime, :d.ndime] = f
                f = fz
        else:
            raise "Field shape '%s' not supported" % str(f.sshape)
        if name not in data[2]:
            data[2][name] = [(time, f)]  # Create point data
        else:
            data[2][name] += [(time, f)]  # Append next time step
    elif f.pshape == (d.nelem, d.etype.ngauss):  # gField
        f1 = f.mean(1)  # Average from Gauss points, better ideas?
        if f.sshape == ():                       # gScalar
            pass
        elif f.sshape == (d.ndime,):             # gVector
            f1 = tools.add_dimensions(f1)
        elif f.sshape == (d.ndime,)*2:           # gTensor
            if d.ndime < 3:
                fz = numpy.zeros((d.nelem, 3, 3))
                fz[..., :d.ndime, :d.ndime] = f1
                f1 = fz
        else:
            raise TypeError("Field shape '%s' not supported" % str(f.sshape))
        etype = _fempy_to_meshio_type[d.etype.name]
        if etype not in data[3]:
            data[3][etype] = {name: [(time, f1)]}
        else:
            if name not in data[3][etype]:
                data[3][etype][name] = [(time, f1)]  # Update cell data
            else:
                data[3][etype][name] += [(time, f1)]  # Append next time step
    return data


#def prepare_time_data(m, name=None, data=None):
#    sol = m.sol
#    d = m.domain
#    nlhs = m.lhs.size
#    y = sol.y.T
#    t = sol.t
#    name = 'lhs' if name is None else name
#    for i in range(len(t)):
#        yi = y[i, :nlhs].reshape(m.lhs.shape)
#        data = prepare_field_data(yi, 'lhs', d, data, t[i])
#    return data


def prepare_model_data(m, what, data=None):
    d = m.domain
    b = d.boundary
    n = True if what in [None, 'all'] else False
    data = prepare_domain_data(d, add_normals=n) if data is None else data
    fields = _get_fields(m, what)
    for name, f in fields:
        if f.domain not in (d, b):  # Skip other domains or subdomains
            continue
        try:
            data = prepare_field_data(f, name, data=data)
        except:
            pass  # Skip silently not supported fields
    return data, len(fields)


def _get_times(m, time):
    if isinstance(time, numpy.ndarray):
        return time  # times have been just prepared
    if time is None:
        t = (m.time,)
    elif isinstance(time, float):
        t = (time,)
    elif isinstance(time, int):
        t = numpy.linspace(m.time0, m.time1, time)
    elif isinstance(time, collections.Iterable):
        if len(time) == 0:
            t = m._sol.t.copy()
        elif len(time) == 3:
            t = numpy.linspace(*time)
        else:
            raise ValueError("Provide empty tuple () or (t0, t1, n)")
    else:
        raise ValueError("Time specification not recognized.")
    return t


def prepare_time_model_data(m, what, time, data=None):
    if not hasattr(m, '_sol'):  # time solution is not available
        return prepare_model_data(m, what, data)
    d = m.domain
    b = d.boundary
    n = True if what in [None, 'all'] else False
    data = prepare_domain_data(d, add_normals=n) if data is None else data
    t0 = m.time  # record model time
    times = _get_times(m, time)
    for t in times:
        m.sol(t)
        fields = _get_fields(m, what)
        for name, f in fields:
            if f.domain not in (d, b):  # Skip other domains or subdomains
                continue
            try:
                data = prepare_field_data(f.copy(), name, data=data, time=t)
            except:
                pass  # Skip silently not supported fields
    m.sol(t0)  # recover initial model time
    return data, len(fields)


def _dump(fname, points, cells, point_data, cell_data, ascii=False):
    if len(points) == 0:
        raise ValueError("Sorry, no data to be processed. Empty domain?")
    write(fname, points, cells,
          point_data=point_data, cell_data=cell_data,
          is_ascii=ascii)


def _preview(points, cells, point_data, cell_data, *gmsh_opts):
    fname = tools.tmpfname('.msh')
    _dump(fname, points, cells, point_data, cell_data)
    out = subprocess.call(['gmsh', fname] + list(gmsh_opts),
                          stderr=subprocess.STDOUT)
    # subprocess.Popen(['gmsh', fname] + list(gmsh_opts), shell=False)
    os.remove(fname)  # remove to not clutter the system
    return out


def preview_domain(d):
    return _preview(*prepare_domain_data(d))


def dump_domain(fname, d, ascii=False):
    return _dump(fname, *(prepare_domain_data(d)+(ascii,)))


def preview_field(f, name=None, d=None):
    if name is None:
        name = f.__class__.__name__
    return _preview(*prepare_field_data(f, name, d))


def dump_field(fname, f, name=None, d=None, ascii=False):
    if name is None:
        name = f.__class__.__name__
    return _dump(fname, *(prepare_field_data(f, name, d)+(ascii,)))


def preview_fields(fields, names=None, d=None):
    data = None
    for i, f in enumerate(fields):
        if names is None:
            name = f.__class__.__name__
        else:
            name = names[i]
        data = prepare_field_data(f, name, d, data)
    return _preview(*data + ('-n',))


def dump_fields(fname, fields, names=None, d=None, ascii=False):
    data = None
    for i, f in enumerate(fields):
        if names is None:
            name = f.__class__.__name__
        else:
            name = names[i]
        data = prepare_field_data(f, name, d, data)
    return _dump(fname, *(data + (ascii,)))


#def preview_time_data(m, name=None):
#    return _preview(*prepare_time_data(m, name))
#
#
#def dump_time_data(fname, m, name=None, ascii=False):
#    return _dump(fname, *(prepare_time_data(m, name)+(ascii,)))


def _get_fields(model, what, blacklist=()):
    if what in [None, 'all']:
        boring = dir(type('dummy', (object,), {}))
        fields = [item
                  for item in inspect.getmembers(model)
                  if item[0] not in boring and
                  not inspect.ismethod(item[1]) and
                  hasattr(item[1], 'domain') and
                  item not in blacklist]
    elif isinstance(what, str) and what != 'all':
        what = [w.strip() for w in what.split(',')]
        fields = [(w, getattr(model, w)) for w in what]
    elif isinstance(what, collections.Iterable):
        fields = [(w, getattr(model, w)) for w in what]
    else:  # it is assumed that what is a field itself
        fields = [(None, what)]
    return fields


def preview_model(m, what='lhs', time=None):
    data, nfields = prepare_time_model_data(m, what, time)
    if nfields > 1:
        data += ('-n',)
    return _preview(*data)


def dump_model(fname, m, what='lhs', time=None, ascii=False):
    data, nfields = prepare_time_model_data(m, what, time)
    return _dump(fname, *(data + (ascii,)))


# Tests
if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')
